# -*- coding: utf-8 -*-
"""Benchmark Markov State Model.

This submodule holds methods for validating Markov state models.

"""
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
import decorit
import numpy as np

from msmhelper import utils
from msmhelper.statetraj import LumpedStateTraj, StateTraj


@decorit.alias('ck_test')
def chapman_kolmogorov_test(trajs, lagtimes, tmax):
    r"""Calculate the Chapman Kolmogorov equation.

    This method estimates the Chapman Kolmogorov equation

    $$T(\tau n) = T^n(\tau)\;.$$

    Projected onto the diagonal this is known as the Chapman Kolmogorov test.
    For more details see, e.g., the review Prinz et al.[^1].

    [^1]: Prinz et al., **Markov models of molecular kinetics: Generation and
        validation**, *J. Chem. Phys.*, 134, 174105 (2011),
        doi:[10.1063/1.3565032](https://doi.org/10.1063/1.3565032)

    Parameters
    ----------
    trajs : StateTraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    lagtimes : list or ndarray int
        Lagtimes for estimating the markov model given in [frames].
    tmax : int
        Longest time to evaluate the CK equation given in [frames].

    Returns
    -------
    cktest : dict
        Dictionary holding for each lagtime the ckequation and with 'md' the
        reference.

    """
    # format input
    trajs = StateTraj(trajs)
    lagtimes = np.atleast_1d(lagtimes)
    lagtimes = np.sort(lagtimes)

    # check that lag times are array of integers
    if not np.issubdtype(lagtimes.dtype, np.integer):
        raise TypeError(
            'Lagtimes needs to be integers but are {0}'.format(lagtimes.dtype),
        )
    if not (lagtimes > 0).all():
        raise TypeError('Lagtimes needs to be positive integers')

    if lagtimes.ndim != 1:
        raise TypeError(
            'Lagtimes needs to be maximal 1d, but {0}'.format(lagtimes),
        )

    if not isinstance(tmax, int) or tmax < 0:
        raise TypeError('tmax needs to be a positive integer')

    ckeqs = {}
    for lagtime in lagtimes:
        ckeqs[lagtime] = _chapman_kolmogorov_test(trajs, lagtime, tmax)
    ckeqs['md'] = _chapman_kolmogorov_test_md(
        trajs, tmin=lagtimes[0], tmax=tmax,
    )

    return ckeqs


def _chapman_kolmogorov_test(trajs, lagtime, tmax):
    r"""Calculate the Chapman Kolmogorov equation $T^n(\tau)$."""
    times = _calc_times(lagtime=lagtime, tmax=tmax)
    ntimes = len(times)
    ckeq = np.empty((trajs.nstates, ntimes))

    # estimate Markov model
    tmat, _ = trajs.estimate_markov_model(lagtime=lagtime)

    is_ergodic = utils.tests.is_ergodic(tmat)
    is_fuzzy_ergodic = utils.tests.is_fuzzy_ergodic(tmat)
    for idx in range(ntimes):
        tmatpow = utils.matrix_power(tmat, idx + 1)
        ckeq[:, idx] = np.diagonal(tmatpow)

    return {
        'ck': {
            state: ckeq[idx]
            for idx, state in enumerate(trajs.states)
        },
        'time': times,
        'is_ergodic': is_ergodic,
        'is_fuzzy_ergodic': is_fuzzy_ergodic,
    }


def _chapman_kolmogorov_test_md(trajs, tmin, tmax, steps=30):
    r"""Calculate the Chapman Kolmogorov equation $T(n\tau)$."""
    times = np.around(np.geomspace(
        start=tmin,
        stop=tmax,
        num=steps,
    )).astype(np.int64)
    # filter duplicated times (for small ranges)
    times = np.unique(times)
    ntimes = len(times)

    ckeq = np.empty((trajs.nstates, ntimes))
    is_ergodic = np.empty(ntimes, dtype=bool)
    is_fuzzy_ergodic = np.empty(ntimes, dtype=bool)

    # enforce using non-lumped trajectory
    if isinstance(trajs, LumpedStateTraj):
        macrotrajs = StateTraj(trajs.trajs)
    else:
        macrotrajs = trajs

    # estimate Markov model
    for idx, time in enumerate(times):
        tmat, _ = macrotrajs.estimate_markov_model(lagtime=time)
        ckeq[:, idx] = np.diagonal(tmat)
        is_ergodic[idx] = utils.tests.is_ergodic(tmat)
        is_fuzzy_ergodic[idx] = utils.tests.is_fuzzy_ergodic(tmat)

    return {
        'ck': {
            state: ckeq[idx]
            for idx, state in enumerate(trajs.states)
        },
        'time': times,
        'is_ergodic': is_ergodic,
        'is_fuzzy_ergodic': is_fuzzy_ergodic,
    }


def _calc_times(lagtime, tmax):
    """Return time array."""
    steps = int(np.floor(tmax / lagtime))
    return lagtime * np.arange(1, steps + 1)



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Set of functions for analyzing the MD trajectory.

This submodule contains methods for estimating various timescales based on a
given state trajectory.

"""
from collections import defaultdict

import decorit
import numba
import numpy as np

from msmhelper.md.comparison import _intersect as intersect
from msmhelper.statetraj import StateTraj


@decorit.alias('estimate_wt')
def estimate_waiting_times(trajs, start, final):
    """Estimates waiting times between stated states.

    The stated states (from/to) will be treated as a basin. The function
    calculates all transitions from first entering the start-basin until first
    reaching the final-basin.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    start : int or list of
        States to start counting.
    final : int or list of
        States to start counting.

    Returns
    -------
    wt : ndarray
        List of waiting times, given in frames.

    """
    # check correct input format
    trajs = StateTraj(trajs)

    states_start, states_final = np.unique(start), np.unique(final)

    if intersect(states_start, states_final):
        raise ValueError('States `start` and `final` do overlap.')

    # check that all states exist in trajectory
    for states in (states_start, states_final):
        if intersect(states, trajs.states) != len(states):
            raise ValueError(
                'Selected states does not exist in state trajectory.',
            )

    # do not convert for pytest coverage
    if numba.config.DISABLE_JIT:
        return _estimate_waiting_times(trajs, states_start, states_final)
    return _estimate_waiting_times(  # pragma: no cover
        numba.typed.List(trajs),
        numba.typed.List(states_start),
        numba.typed.List(states_final),
    )


@numba.njit
def _estimate_waiting_times(trajs, states_start, states_final):
    """Estimate waiting time between subsets of states."""
    times = []
    for traj in trajs:
        times.extend(
            _estimate_waiting_times_singletraj(
                traj, states_start, states_final,
            ),
        )
    return np.array(times)


@numba.njit
def _estimate_waiting_times_singletraj(traj, states_start, states_final):
    """Estimate waiting time between subsets of states for a single traj."""
    paths_idx = _estimate_events_singletraj(traj, states_start, states_final)
    return [idx_end - idx_start for idx_start, idx_end in paths_idx]


def estimate_paths(trajs, start, final):
    """Estimates paths and waiting times between stated states.

    The stated states (from/to) will be treated as a basin. The function
    calculates all transitions from first entering the start-basin until first
    reaching the final-basin. The results will be listed by the corresponding
    pathways, where loops are removed occuring first.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    start : int or list of
        States to start counting.
    final : int or list of
        States to start counting.

    Returns
    -------
    paths : dict
        Dictionary containing the the paths as keys and and an array holding
        the times of all paths as value.

    """
    # check correct input format
    trajs = StateTraj(trajs)

    states_start, states_final = np.unique(start), np.unique(final)

    if intersect(states_start, states_final):
        raise ValueError('States `start` and `final` do overlap.')

    # check that all states exist in trajectory
    for states in (states_start, states_final):
        if intersect(states, trajs.states) != len(states):
            raise ValueError(
                'Selected states does not exist in state trajectory.',
            )

    # do not convert for pytest coverage
    if numba.config.DISABLE_JIT:  # pragma: no cover
        path_tuples = _estimate_paths(trajs, states_start, states_final)
    else:
        path_tuples = _estimate_paths(  # pragma: no cover
            numba.typed.List(trajs),
            numba.typed.List(states_start),
            numba.typed.List(states_final),
        )
    paths = defaultdict(list)
    for path, time in path_tuples:
        paths[tuple(path)].append(time)

    return paths


@numba.njit
def _estimate_paths(trajs, states_start, states_final):
    """Estimate waiting time between subsets of states."""
    paths = []
    for traj in trajs:
        paths.extend(
            _estimate_paths_singletraj(
                traj, states_start, states_final,
            ),
        )
    return paths


@numba.njit
def _estimate_events_singletraj(traj, states_start, states_final):
    """Estimate waiting time between subsets of states for a single traj."""
    paths_idx = []
    propagates_forwards = False
    idx_start = 0
    for idx in range(len(traj)):  # noqa: WPS518
        state = traj[idx]

        if not propagates_forwards and state in states_start:
            propagates_forwards = True
            idx_start = idx
        elif propagates_forwards and state in states_final:
            propagates_forwards = False
            paths_idx.append((idx_start, idx))

    return paths_idx


@numba.njit
def _estimate_paths_singletraj(traj, states_start, states_final):
    """Estimate waiting time between subsets of states for a single traj."""
    paths_idx = _estimate_events_singletraj(traj, states_start, states_final)

    paths = []
    for idx_start, idx_end in paths_idx:
        for state in traj[idx_start: idx_end + 1]:  # noqa: WPS518
            if state in states_start:
                path = [int(define_type_var) for define_type_var in range(0)]
            elif state in path:
                path = path[: path.index(state)]
            path.append(state)

        paths.append((path, idx_end - idx_start))
    return paths



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""# Input and Output Text Files

This submodule contains all methods related to reading data from text files and
writing data to text files, including helpful header comments.

"""
import datetime
import getpass  # noqa: SC100 # get user name with getpass.getuser()
import os
import platform  # get pc name with platform.node()
import sys

import numpy as np
import pandas as pd

import __main__ as main
from msmhelper import utils


class FileError(Exception):
    """An exception for wrongly formated input files."""


def opentxt(file_name, comment='#', nrows=None, **kwargs):
    r"""Open a text file.

    This method can load an nxm array of floats from an ascii file. It uses
    either pandas read_csv for a single comment or as fallback the slower
    [np.loadtxt][numpy.loadtxt] for multiple comments.

    !!! warning
        In contrast to pandas the order of usecols will be used. So if
        using `data = opentxt(..., uscols=[1, 0])` you access the first column
        by `data[:, 0]` and the second one by `data[:, 1]`.

    Parameters
    ----------
    file_name : string
        Name of file to be opened.
    comment : str or array of str, optional
        Characters with which a comment starts.
    nrows : int, optional
        The maximum number of lines to be read
    usecols : int-array, optional
        Columns to be read from the file (zero indexed).
    skiprows : int, optional
        The number of leading rows which will be skipped.
    dtype : data-type, optional
        Data-type of the resulting array. Default: float.

    Returns
    -------
    array : ndarray
        Data read from the text file.

    """
    if len(comment) == 1:
        # pandas does not support array of single char
        if not isinstance(comment, str):
            comment = comment[0]

        # force pandas to load in stated order without sorting
        cols = kwargs.pop('usecols', None)
        if cols is not None:
            idx = np.argsort(cols)
            cols = np.atleast_1d(cols).astype(np.int32)[idx]

        array = pd.read_csv(
            file_name,
            sep=r'\s+',
            header=None,
            comment=comment,
            nrows=nrows,
            usecols=cols,
            **kwargs,
        ).values

        if array.shape[-1] == 1:
            array = array.flatten()
        # swap columns back to ensure correct order
        elif cols is not None:
            array = utils.swapcols(array, idx, np.arange(len(idx)))

        return array

    return np.loadtxt(
        file_name,
        comments=comment,
        max_rows=nrows,
        **kwargs,
    )


def savetxt(file_name, array, header=None, fmt='%.5f'):  # noqa: WPS323
    """Save nxm array of floats to a text file.

    It uses numpys savetxt method and extends the header with information
    of execution.

    Parameters
    ----------
    file_name : string
        File name to store data.
    array : ndarray
        Data to be stored.
    header : str, optional
        Comment written into the header of the output file.
    fmt : str or sequence of strs, optional
        See [numpy.savetxt][].

    """
    # prepare header comments
    RUI = _get_runtime_user_information()

    header_comment = (
        'This file was generated by {script_dir}/{script_name}:\n{args}' +
        '\n\n{date}, {user}@{pc}'
    ).format(**RUI, args=' '.join(sys.argv))

    if header:  # print column title if given
        header_comment += '\n{0}'.format(header)

    # save file
    np.savetxt(file_name, array, fmt=fmt, header=header_comment)


def opentxt_limits(file_name, limits_file=None, **kwargs):
    """Load file and split according to limit file.

    If limits_file is not provided it will return `[traj]`.

    Parameters
    ----------
    file_name : string
        Name of file to be opened.
    limits_file : str, optional
        File name of limit file. Should be single column ascii file.
    **kwargs
        See parameters defined in [opentxt][msmhelper.io.opentxt]

    Returns
    -------
    traj : ndarray
        Return array of subtrajectories.

    """
    # open trajectory
    traj = opentxt(file_name, **kwargs)

    # open limits
    limits = open_limits(limits_file=limits_file, data_length=len(traj))

    # split trajectory
    return np.split(traj, limits)[:-1]


def openmicrostates(file_name, limits_file=None, **kwargs):
    """Load 1d file and split according to limit file.

    Both, the limit file and the trajectory file needs to be a single column
    file. If limits_file is not provided it will return [traj]. The trajectory
    will of dtype np.int16, so the states needs to be smaller than 32767.

    Parameters
    ----------
    file_name : string
        Name of file to be opened.
    limits_file : str, optional
        File name of limit file. Should be single column ascii file.
    **kwargs
        See parameters defined in [opentxt][msmhelper.io.opentxt]

    Returns
    -------
    traj : ndarray
        Return array of subtrajectories.

    """
    # open trajectory
    if 'dtype' in kwargs and not np.issubdtype(kwargs['dtype'], np.integer):
        raise TypeError('dtype should be integer')
    else:
        kwargs['dtype'] = np.int16

    # load split trajectory
    traj = opentxt_limits(file_name, limits_file, **kwargs)

    if len(traj[0].shape) != 1:
        raise FileError('Microstate trjectory shoud be single column file.')

    return traj


def open_limits(data_length, limits_file=None):
    """Load and check limit file.

    The limits give the length of each single trajectory. So e.g.
    [5, 5, 5] for 3 equally-sized subtrajectories of length 5.

    Parameters
    ----------
    data_length : int
        Length of data read.
    limits_file : str, optional
        File name of limit file. Should be single column ascii file.

    Returns
    -------
    limits : ndarray
        Return cumsum of limits.

    """
    if limits_file is None:
        return np.array([data_length])  # for single trajectory

    # open limits file
    limits = opentxt(limits_file)
    if len(limits.shape) != 1:
        raise FileError('Shoud be single column file.')

    # convert to cumulative sum
    limits = np.cumsum(limits)
    if data_length != limits[-1]:
        raise ValueError('Limits are inconsistent to data.')

    return limits


def _get_runtime_user_information():
    r"""Get user runtime information.

    !!! warning
        For python 3.5 or lower the date is not formatted and contains
        microscends.

    Returns
    -------
    RUI : dict
        Holding username in 'user', pc name in 'pc', date of execution 'date',
        path of execution 'script_dir' and name of execution main file
        'script_name'. In case of interactive usage, script_name is 'console'.

    """
    try:
        script_dir, script_name = os.path.split(
            os.path.abspath(main.__file__),  # noqa: WPS609
        )
    except AttributeError:  # pragma: no cover
        script_dir, script_name = '', 'console'

    # get time without microseconds
    date = datetime.datetime.now()
    if sys.version_info >= (3, 6):
        date = date.isoformat(sep=' ', timespec='seconds')

    return {
        'user': getpass.getuser(),
        'pc': platform.node(),
        'date': date,
        'script_dir': script_dir,
        'script_name': script_name,
    }



# -*- coding: utf-8 -*-
"""# Utils

This submodule provides utility functions that can be used to manipulate and test data, such as filtering and validation methods. The functions in this submodule can be used in conjunction with other parts of the software to perform a variety of tasks, making it an essential part of the package.

The submodule is structured into the following submodules:

- [**datasets:**][msmhelper.utils.datasets] This submodule contains all methods related to create example datasets. This submodule needs to be imported explicitly!
- [**filtering:**][msmhelper.utils.filtering] This submodule contains all methods related to dynamical smoothening.
- [**tests:**][msmhelper.utils.tests] This submodule holds functions to tests for given properties, e.g., if a matrix is ergodic, quadratic, etc.


"""
__all__ = [
    'find_first',
    'format_state_traj',
    'matrix_power',
    'rename_by_index',
    'rename_by_population',
    'runningmean',
    'shift_data',
    'swapcols',
    'unique',
]

from . import filtering, tests
from ._utils import (
    find_first,
    format_state_traj,
    matrix_power,
    rename_by_index,
    rename_by_population,
    runningmean,
    shift_data,
    swapcols,
    unique,
)



# -*- coding: utf-8 -*-
"""# Matplotlib-Based Plotting Routines

This submodule contains methods related to visualize some of the standard Markov state plots.

"""
__all__ = [
    'plot_ck_test',
    'plot_wtd',
]

from ._ck_test import plot_ck_test
from ._wtd import plot_wtd



# -*- coding: utf-8 -*-
""" --8<-- "docs/tutorials/msmhelper.md" """
__all__ = [
    'opentxt',
    'savetxt',
    'opentxt_limits',
    'openmicrostates',
    'open_limits',
    'LumpedStateTraj',
    'StateTraj',
    'rename_by_population',
    'rename_by_index',
    'shift_data',
    'unique',
]

from . import md, msm, plot, utils
from .io import (
    opentxt,
    savetxt,
    opentxt_limits,
    openmicrostates,
    open_limits,
)
from .statetraj import LumpedStateTraj, StateTraj
from .utils import (
    rename_by_population,
    rename_by_index,
    shift_data,
    unique,
)

__version__ = '1.0.3'



# -*- coding: utf-8 -*-
"""# Time Series Analysis

This submodule offers techniques for the analysis of state trajectories&mdash;commonly known as Molecular Dynamics (MD)&mdash;without relying on Markov state models. It encompasses functions for determining timescales, recognizing significant events, correcting dynamical anomalies, and evaluating various state discretization methods.  These functions provide a comprehensive solution for analyzing time-series data and understanding the underlying dynamics of complex systems.

The submodule is structured into the following submodules:

- [**comparison:**][msmhelper.md.comparison] This submodule holds methods to quantify the similarity of different state discretizations.
- [**corrections:**][msmhelper.md.corrections] This submodule provides an implementation of dynamical coring.
- [**timescales:**][msmhelper.msm.timescales] This submodule contains methods for estimating various timescales based on discrete time series.

"""
__all__ = [
    'compare_discretization',
    'dynamical_coring',
    'estimate_waiting_times',
    'estimate_wt',
    'estimate_paths',
]

from .comparison import compare_discretization
from .corrections import dynamical_coring
from .timescales import (
    estimate_waiting_times,
    estimate_wt,
    estimate_paths,
)



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Plot the Chapman-Kolmogorov test."""
from os.path import splitext

import click
import msmhelper as mh
import prettypyplot as pplt
from matplotlib import pyplot as plt


@click.command(no_args_is_help='-h')
@click.option(
    '--filename',
    '-f',
    required=True,
    type=click.Path(exists=True),
    help='Path to state trajectory file (single column ascii file).',
)
@click.option(
    '--microfilename',
    required=False,
    type=click.Path(exists=True),
    help=(
        'Path to microstate trajectory file (single column ascii file) to use '
        'Hummer-Szabo projection.'
    ),
)
@click.option(
    '--concat-limits',
    '-c',
    required=False,
    type=click.Path(exists=True),
    help=(
        'Path to concatination limit file given the length of all ' +
        r'trajectories, e.g. "3\n3\n5"'
    ),
)
@click.option(
    '--output',
    '-o',
    type=click.Path(),
    help=(
        'Output basename of figure. Needs to have a valid extension (".pdf", '
        '".svg", ".png"). Default format is pdf.'
    ),
)
@click.option(
    '--lagtimes',
    required=True,
    nargs=5,
    type=click.IntRange(min=1),
    help='5 (!) Lag times given in frames to estimate Markov state model.',
)
@click.option(
    '--frames-per-unit',
    required=True,
    type=click.FLOAT,
    help='Number of frames per unit.',
)
@click.option(
    '--unit',
    required=True,
    type=click.Choice(
        ['fs', 'ps', 'ns', 'us', 'frames'],
        case_sensitive=False,
    ),
    help='Unit of data.',
)
@click.option(
    '--grid',
    type=click.IntRange(min=1),
    nargs=2,
    default=(1, 1),
    help='Number of rows and cols.',
)
@click.option(
    '--max-time',
    type=click.IntRange(min=1),
    default=int(10**4),
    help='Largest time value to evaluate and plot the test.',
)
def ck_test(
    filename,
    microfilename,
    concat_limits,
    output,
    lagtimes,
    frames_per_unit,
    unit,
    grid,
    max_time,
):
    """Estimation and visualization of the Chapman-Kolmogorov test."""
    # setup matplotlib
    pplt.use_style(
        figsize=0.8, true_black=True, colors='pastel_autunm', latex=False,
    )

    # load file
    trajs = mh.openmicrostates(filename, limits_file=concat_limits)
    if microfilename:
        microtrajs = mh.openmicrostates(
            microfilename, limits_file=concat_limits,
        )
        trajs = mh.LumpedStateTraj(trajs, microtrajs)
    else:
        trajs = mh.StateTraj(trajs)

    # perform test
    ck = mh.msm.ck_test(trajs, lagtimes=lagtimes, tmax=max_time)

    # plot result
    nrows, ncols = grid
    for chunk in mh.plot._ck_test._split_array(trajs.states, nrows * ncols):
        mh.plot.plot_ck_test(
            ck=ck,
            states=chunk,
            frames_per_unit=frames_per_unit,
            unit=unit,
            grid=grid,
        )

        # save figure and continue
        if output is None:
            basename = f'{filename}.sh' if microfilename else filename
            output = f'{basename}.cktest.pdf'
        # insert state_str between pathname and extension
        path, ext = splitext(output)
        pplt.savefig(f'{path}.state{chunk[0]:.0f}-{chunk[-1]:.0f}{ext}')
        plt.close()


if __name__ == '__main__':
    ck_test()  # pragma: no cover



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Set of datasets to use for the tutorials."""
import functools

import numpy as np

from msmhelper.msm.timescales import _propagate_MCMC
from .tests import is_transition_matrix
from ._utils import shift_data


def _decorate_tmat(tmat):
    """Add tmat as attribute of function."""
    def decorator_tmat(func):
        @functools.wraps(func)
        def wrapper_tmat(*args, **kwargs):
            return func(*args, **kwargs)

        wrapper_tmat.tmat = tmat
        return wrapper_tmat

    return decorator_tmat


def propagate_tmat(tmat, nsteps, start=None):
    """Markov chain Monte Carlo propagation of transition matrix for nsteps.

    Parameters
    ----------
    tmat : ndarray
        Transition matrix to propagate.
    nsteps : int
        Number of steps to propagate.
    start : int
        Index where to start. If `None` a random number will be used.

    Returns
    -------
    traj : ndarray
        Markov chain Monte Carlo state trajectory of given tmat.

    """
    if not is_transition_matrix(tmat):
        raise ValueError('tmat needs to be a row-normalized matrix.')

    n_states = len(tmat)
    cummat = np.cumsum(tmat, axis=1)
    cummat[:, -1] = 1  # enforce exact normalization
    cummat_perm = np.tile(np.arange(n_states), (n_states, 1))

    if start is None:
        start = np.random.randint(n_states)

    return _propagate_MCMC(
        cummat=(cummat, cummat_perm),
        start=start,
        steps=nsteps,
    )


def _hummer15_nstate(
    *, n_states, rate_k, rate_h, nsteps, return_macrotraj=False
):
    """N-state model inspired by Hummer and Szabo 15.

    Gerhard Hummer and Attila Szabo
    The Journal of Physical Chemistry B 2015 119 (29), 9029-9037
    DOI: [10.1021/jp508375q](https://pubs.acs.org/doi/10.1021/jp508375q)

    Parameters
    ----------
    rate_k : float
        Rate between state 1<->2, 3<->4, etc.
    rate_h : float
        Rate between state 2<->3, 4<->5, etc.
    nsteps : int
        Number of steps to propagate.
    return_macrotraj : bool, optional
        If `True` return a macrotraj where state (1,2), (3,4), etc. are lumped
        as well.

    Returns
    -------
    traj : ndarray
        Markov chain Monte Carlo state trajectory.
    macrotraj : ndarray
        Markov chain Monte Carlo macrostate trajectory if `macrotraj=True`.

    """
    tmat = np.zeros((n_states, n_states))
    for state in range(n_states):
        for neighbor, rate in (
            (state - 1, rate_k if state % 2 else rate_h),
            (state + 1, rate_h if state % 2 else rate_k),
        ):
            if 0 <= neighbor < n_states:
                tmat[state, neighbor] = rate

    # set diagonal
    tmat[np.diag_indices_from(tmat)] = 1 - np.sum(tmat, axis=1)

    microstates = np.arange(1, n_states + 1)
    microtraj = shift_data(
        propagate_tmat(tmat, nsteps),
        np.arange(n_states),
        microstates,
    )

    if return_macrotraj:
        macrostates = np.tile(
            np.arange(1, n_states // 2 + 1), (2, 1),
        ).T.flatten()
        return (
            microtraj,
            shift_data(
                microtraj,
                microstates,
                macrostates,
            ),
        )
    return microtraj


def hummer15_4state(rate_k, rate_h, nsteps, return_macrotraj=False):
    """Four state model taken from Hummer and Szabo 15.

    Gerhard Hummer and Attila Szabo
    The Journal of Physical Chemistry B 2015 119 (29), 9029-9037
    DOI: [10.1021/jp508375q](https://pubs.acs.org/doi/10.1021/jp508375q)

    Parameters
    ----------
    rate_k : float
        Rate between state 1<->2 and 3<->4.
    rate_h : float
        Rate between state 2<->3.
    nsteps : int
        Number of steps to propagate.
    return_macrotraj : bool, optional
        If `True` return a macrotraj where state (1,2) and (3,4) are lumped
        as well.

    Returns
    -------
    traj : ndarray
        Markov chain Monte Carlo state trajectory.
    macrotraj : ndarray
        Markov chain Monte Carlo macrostate trajectory if `macrotraj=True`.

    """
    return _hummer15_nstate(
        n_states=4,
        rate_k=rate_k,
        rate_h=rate_h,
        nsteps=nsteps,
        return_macrotraj=return_macrotraj,
    )


def hummer15_8state(rate_k, rate_h, nsteps, return_macrotraj=False):
    """Eight state model inspired by Hummer and Szabo 15.

    Gerhard Hummer and Attila Szabo
    The Journal of Physical Chemistry B 2015 119 (29), 9029-9037
    DOI: [10.1021/jp508375q](https://pubs.acs.org/doi/10.1021/jp508375q)

    Parameters
    ----------
    rate_k : float
        Rate between state 1<->2, 3<->4, 5<->6, 7<->8.
    rate_h : float
        Rate between state 2<->3, 4<->5, 6<->7.
    nsteps : int
        Number of steps to propagate.
    return_macrotraj : bool, optional
        If `True` return a macrotraj where state (1,2), (3,4), (5,6), and (7,8)
        are lumped as well.

    Returns
    -------
    traj : ndarray
        Markov chain Monte Carlo state trajectory.
    macrotraj : ndarray
        Markov chain Monte Carlo macrostate trajectory if `macrotraj=True`.

    """
    return _hummer15_nstate(
        n_states=8,
        rate_k=rate_k,
        rate_h=rate_h,
        nsteps=nsteps,
        return_macrotraj=return_macrotraj,
    )


_NAGEL20_4STATE_TMAT = np.array([
    [0.92, 0.04, 0.04, 0.00],
    [0.10, 0.70, 0.10, 0.10],
    [0.10, 0.10, 0.70, 0.10],
    [0.00, 0.04, 0.04, 0.92],
])


@_decorate_tmat(_NAGEL20_4STATE_TMAT)
def nagel20_4state(nsteps):
    """Four state model taken from Nagel et al. 20.

    Daniel Nagel, Anna Weber, and Gerhard Stock
    Journal of Chemical Theory and Computation 2020 16 (12), 7874-7882
    DOI: [10.1021/acs.jctc.0c00774](https://doi.org/10.1021/acs.jctc.0c00774)

    Parameters
    ----------
    nsteps : int
        Number of steps to propagate.

    Returns
    -------
    traj : ndarray
        Markov chain Monte Carlo state trajectory.

    """
    return propagate_tmat(_NAGEL20_4STATE_TMAT, nsteps) + 1


_NAGEL20_6STATE_TMAT = np.array([
    [0.90, 0.03, 0.07, 0.00, 0.00, 0.00],
    [0.05, 0.63, 0.15, 0.07, 0.10, 0.00],
    [0.08, 0.04, 0.60, 0.14, 0.14, 0.00],
    [0.00, 0.12, 0.15, 0.50, 0.15, 0.08],
    [0.00, 0.04, 0.16, 0.10, 0.60, 0.10],
    [0.00, 0.00, 0.00, 0.08, 0.02, 0.90],
])


@_decorate_tmat(_NAGEL20_6STATE_TMAT)
def nagel20_6state(nsteps):
    """Six state model taken from Nagel et al. 20.

    Daniel Nagel, Anna Weber, and Gerhard Stock
    Journal of Chemical Theory and Computation 2020 16 (12), 7874-7882
    DOI: [10.1021/acs.jctc.0c00774](https://doi.org/10.1021/acs.jctc.0c00774)

    Parameters
    ----------
    nsteps : int
        Number of steps to propagate.

    Returns
    -------
    traj : ndarray
        Markov chain Monte Carlo state trajectory.

    """
    return propagate_tmat(_NAGEL20_6STATE_TMAT, nsteps) + 1



# -*- coding: utf-8 -*-
"""Tests for the msm module.

BSD 3-Clause License
Copyright (c) 2019-2023, Daniel Nagel
All rights reserved.

"""
import random

import numpy as np
import pytest

from msmhelper import LumpedStateTraj
from msmhelper.msm import timescales


@pytest.fixture
def rand():
    random.seed(43)
    np.random.seed(43)


@pytest.mark.parametrize('transmat, lagtime, result', [
    (
        [[0.8, 0.2, 0.0], [0.2, 0.78, 0.02], [0.0, 0.2, 0.8]],
        2,
        -2 / np.log([4 / 5, 29 / 50]),
    ),
    (
        [[0.1, 0.9], [0.8, 0.2]],
        1,
        [np.nan],
    ),
])
def test__implied_timescales(transmat, lagtime, result):
    """Test implied timescale."""
    ntimescales = len(result)
    impl = timescales._implied_timescales(
        transmat, lagtime, ntimescales=ntimescales,
    )
    np.testing.assert_array_almost_equal(impl, result)


@pytest.mark.parametrize('trajs, lagtimes, kwargs, result, error', [
    (
        [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
        [1, 2],
        {},
        [-1 / np.log([2 / 3]), -2 / np.log([7 / 15])],
        None,
    ),
    (
        [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
        [1, 2],
        {'ntimescales': 1},
        [-1 / np.log([2 / 3]), -2 / np.log([7 / 15])],
        None,
    ),
    (
        [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
        [-1, 2],
        {},
        None,
        TypeError,
    ),
    (
        [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
        [1, 2.3],
        {},
        None,
        TypeError,
    ),
    (
        [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1],
        [1, 2],
        {'reversible': True},
        None,
        NotImplementedError,
    ),
])
def test_implied_timescales(trajs, lagtimes, kwargs, result, error):
    """Test implied timescale."""
    if error is None:
        impl = timescales.implied_timescales(trajs, lagtimes, **kwargs)
        np.testing.assert_array_almost_equal(impl, result)
    else:
        with pytest.raises(error):
            timescales.implied_timescales(trajs, lagtimes, **kwargs)


@pytest.mark.parametrize('traj, lagtime, result, error', [
    (
        [1, 2, 1, 2, 2, 1, 1, 2, 2, 1, 2],
        1,
        (
            [[0.8, 1.0], [0.6, 1.0]],
            [[1, 0], [0, 1]],
        ),
        None,
    ),
    (
        LumpedStateTraj(
            [3, 1, 1, 1, 1, 2, 1, 2, 1, 1, 1, 1, 2, 2, 3],
            [4, 1, 1, 1, 3, 2, 3, 2, 1, 1, 1, 3, 2, 2, 4],
        ),
        1,
        None,
        ValueError,
    ),
])
def test__get_cummat(traj, lagtime, result, error):
    """Test cumulative matrix."""
    if error is None:
        cummat, perm = timescales._get_cummat(traj, lagtime)
        np.testing.assert_array_almost_equal(cummat, result[0])
        np.testing.assert_array_almost_equal(perm, result[1])
    else:
        with pytest.raises(error):
            timescales._get_cummat(traj, lagtime)


@pytest.mark.parametrize('cummat, idx_from, result', [
    (
        (
            np.array([[0.8, 1.0], [0.6, 1.0]]),
            np.array([[1, 0], [0, 1]]),
        ),
        0,
        1,
    ),
    (
        (
            np.array([[0.8, 1.0], [0.6, 1.0]]),
            np.array([[1, 0], [0, 1]]),
        ),
        1,
        0,
    ),
])
def test__propagate_MCMC_step(cummat, idx_from, result, rand):
    """Test MCMC propagation step."""
    idx_next = timescales._propagate_MCMC_step(cummat, idx_from)
    assert idx_next == result


@pytest.mark.parametrize('cummat, start, steps, result_pop', [
    (
        (
            np.array([[0.8, 1.0], [0.8, 1.0]]),
            np.array([[0, 1], [1, 0]]),
        ),
        0,
        10000,
        {0: 0.5, 1: 0.5},
    ),
    (
        (
            np.array([[0.9, 1.0], [0.8, 1.0]]),
            np.array([[0, 1], [1, 0]]),
        ),
        0,
        10000,
        {0: 0.6667, 1: 0.3333},
    ),
])
def test__propagate_MCMC(cummat, start, steps, result_pop, rand):
    """Test MCMC propagation step."""
    mcmc = timescales._propagate_MCMC(cummat, start, steps)
    # check if population is approxematly the given one:
    for state, pop in result_pop.items():
        np.testing.assert_approx_equal(
            len(mcmc[mcmc == state]) / steps,
            pop,
            significant=2,
        )


@pytest.mark.parametrize('trajs, lagtime, steps, start, error', [
    (
        [1, 1, 1, 2, 2, 1, 1],
        1,
        10,
        1,
        None,
    ),
    (
        [1, 1, 1, 2, 2, 1, 1],
        1,
        10,
        -1,
        None,
    ),
    (
        [1, 1, 1, 2, 2, 1, 1],
        1,
        10,
        3,
        ValueError,
    ),
])
def test_propagate_MCMC(trajs, lagtime, steps, start, error, rand):
    """Test MCMC propagation."""
    if error is None:
        mcmc = timescales.propagate_MCMC(trajs, lagtime, steps, start)

        # due to different random numbers on local machine and github
        # workflows compare only basic stats
        np.testing.assert_array_almost_equal(
            np.unique(mcmc), np.unique(trajs),
        )
    else:
        with pytest.raises(error):
            timescales.propagate_MCMC(trajs, lagtime, steps, start)


@pytest.mark.parametrize(
    'trajs, lagtime, start, final, return_list, result, error',
    [
        (
            [1, 1, 1, 2, 2, 1, 1],
            1,
            1,
            2,
            False,
            (np.array([0, 10, 3, 0, 1]) / 14, [0, 1, 2, 3, 4]),
            None,
        ),
        (
            [1, 1, 1, 2, 2, 1, 1],
            2,
            1,
            2,
            False,
            (np.array([0, 10, 3, 0, 1]) / 14 / 2, [0, 2, 4, 6, 8]),
            None,
        ),
        (
            [1, 1, 1, 2, 2, 1, 1],
            1,
            1,
            2,
            True,
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 4],
            None,
        ),
        (
            [1, 1, 1, 2, 2, 1, 1],
            2,
            1,
            2,
            True,
            [2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 4, 4, 4, 8],
            None,
        ),
        (
            [1, 1, 1, 2, 2, 1, 1],
            1,
            1,
            1,
            None,
            None,
            ValueError,
        ),
        (
            [1, 1, 1, 2, 2, 1, 1],
            1,
            1,
            4,
            None,
            None,
            ValueError,
        ),
    ],
)
def test__estimate_time(
        trajs, lagtime, start, final, return_list, result, error,
):
    """Test WT/TT time wrapper."""
    # define deterministic estimator to debug wrapper only
    def estimator(cummat, start, states_from, states_to, steps):
        return {1: 10, 2: 3, 4: 1}

    # not used by estimator
    steps = 10

    if error is None:
        ts = timescales._estimate_times(
            trajs=trajs,
            lagtime=lagtime,
            start=start,
            final=final,
            steps=steps,
            estimator=estimator,
            return_list=return_list,
        )
        if isinstance(ts, tuple):
            for idx in (0, 1):
                np.testing.assert_array_almost_equal(ts[0], result[0])
        else:
            np.testing.assert_array_almost_equal(ts, result)
    else:
        with pytest.raises(error):
            timescales._estimate_times(
                trajs=trajs,
                lagtime=lagtime,
                start=start,
                final=final,
                steps=steps,
                estimator=estimator,
                return_list=return_list,
            )



# -*- coding: utf-8 -*-
"""Tests for the plot module.

BSD 3-Clause License
Copyright (c) 2019-2023, Daniel Nagel
All rights reserved.

"""
import numpy as np
import pytest
import prettypyplot as pplt
from matplotlib import pyplot as plt

import msmhelper as mh
from msmhelper.utils.datasets import hummer15_8state


@pytest.mark.mpl_image_compare(remove_text=True)
@pytest.mark.parametrize('trajs, kwargs', [
    (hummer15_8state(0.2, 0.05, 100000), {'grid': (4, 2)}),
    (hummer15_8state(0.2, 0.05, 100000), {'grid': (2, 4)}),
    (hummer15_8state(0.2, 0.05, 100000), {'grid': (1, 8)}),
])
def test_plot_ck_test(trajs, kwargs):
    """Test plotting ck_test."""
    lagtimes = [1, 2, 3, 4, 5]
    tmax = 500
    ck = mh.msm.ck_test(trajs, lagtimes=lagtimes, tmax=tmax)
    pplt.use_style(latex=False)
    return mh.plot.plot_ck_test(ck, **kwargs)


@pytest.fixture
def wtd():
    """Define wtd object."""
    lagtimes = np.arange(1, 25)

    wtd = {
        time: {
            'whislo': np.log(time + 1) - 3,
            'q1': np.log(time + 1) - 1,
            'med': np.log(time + 1),
            'q3': np.log(time + 1) + 1,
            'whishi': np.log(time + 1) + 3,
            'fliers': np.linspace(
                np.log(time + 1) + 3,
                np.log(time + 1) + 6,
                5,
            ),
        }
        for time in lagtimes
    }

    # add MD
    wtd['MD'] = [wtd[8]]
    return wtd


@pytest.mark.mpl_image_compare(remove_text=True)
@pytest.mark.parametrize('kwargs', [
    {},
    {'show_md': False},
    {'show_fliers': True},
    {'show_md': False, 'show_fliers': True},
])
def test_plot_wtd(wtd, kwargs):
    """Test plotting ck_test."""
    pplt.use_style(latex=False)
    fig, ax = plt.subplots()
    mh.plot.plot_wtd(wtd=wtd, ax=ax, **kwargs)
    return fig



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Gaussian filtering of coordinates."""
import click
import numpy as np
import msmhelper as mh


@click.command(
    no_args_is_help=True,
)
@click.option(
    '-i',
    '--input',
    'input_file',
    required=True,
    type=click.Path(exists=True),
    help=(
        'Path to input file. Needs to be of shape (n_samples, n_features).'
        ' All comment lines need to start with "#"'
    ),
)
@click.option(
    '--concat-limits',
    '-c',
    required=False,
    type=click.Path(exists=True),
    help=(
        'Path to concatination limit file given the length of all ' +
        r'trajectories, e.g. "3\n3\n5"'
    ),
)
@click.option(
    '-s',
    '--sigma',
    required=True,
    type=click.FloatRange(min=1),
    help='Standard deviation given in [frames] used for Gaussian kernel',
)
@click.option(
    '-o',
    '--output',
    'output_file',
    type=click.Path(),
    help='Path to output file, will be of same shape as input',
)
def gaussian_filtering(input_file, concat_limits, sigma, output_file):
    """Applying gaussian filter on time series."""
    if not output_file:
        output_file = f'{input_file}.gaussian{sigma:.0f}f'
    data = mh.opentxt_limits(
        input_file, limits_file=concat_limits, dtype=np.float32,
    )
    if data[0].ndim == 1:
        data = [part.reshape(-1, 1) for part in data]

    filtered_data = [
        mh.utils.filtering.gaussian_filter(part, sigma=sigma)
        for part in data
    ]

    mh.savetxt(
        output_file,
        np.vstack(filtered_data),
        header=f'Gaussian filtered along time with std={sigma:.3f} frames',
        fmt='%.5f',
    )


if __name__ == '__main__':
    gaussian_filtering()  # pragma: no cover



# -*- coding: utf-8 -*-
"""Create Markov State Model.

This submodule contains all methods related to estimate the Markov state model.

"""
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.

import decorit
import numba
import numpy as np

from msmhelper.utils import tests
from msmhelper.msm.utils import linalg
from msmhelper.statetraj import StateTraj


def estimate_markov_model(trajs, lagtime):
    """Estimates Markov State Model.

    This method estimates the MSM based on the transition count matrix.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    lagtime : int
        Lag time for estimating the markov model given in [frames].

    Returns
    -------
    T : ndarray
        Transition rate matrix.
    permutation : ndarray
        Array with corresponding states.

    """
    trajs = StateTraj(trajs)
    return trajs.estimate_markov_model(lagtime)


def _estimate_markov_model(trajs, lagtime, nstates, perm=None):
    """Estimate MSM based on the transition count matrix."""
    # convert trajs to numba list # noqa: SC100
    if not numba.config.DISABLE_JIT:  # pragma: no cover
        trajs = numba.typed.List(trajs)

    if perm is None:
        perm = np.arange(nstates)

    Tcount = _generate_transition_count_matrix(trajs, lagtime, nstates)
    return row_normalize_matrix(Tcount), perm


@numba.njit
def _generate_transition_count_matrix(trajs, lagtime, nstates):
    """Generate a simple transition count matrix from multiple trajectories."""
    # initialize matrix
    T_count = np.zeros((nstates, nstates), dtype=np.int64)

    for traj in trajs:
        for stateFrom, stateTo in zip(traj[:-lagtime], traj[lagtime:]):
            T_count[stateFrom, stateTo] += 1

    return T_count


@numba.njit
def row_normalize_matrix(mat):
    """Row normalize the given 2d matrix.

    Parameters
    ----------
    mat : ndarray
        Matrix to be row normalized.

    Returns
    -------
    mat : ndarray
        Normalized matrix.

    """
    row_sum = np.sum(mat, axis=1)
    if not row_sum.all():
        row_sum[row_sum == 0] = 1

    # due to missing np.newaxis row_sum[:, np.newaxis] becomes # noqa: SC100
    return mat / row_sum.reshape(mat.shape[0], 1)


@decorit.alias('peq')
def equilibrium_population(tmat, allow_non_ergodic=True):
    """Calculate equilibirum population.

    If there are non ergodic states, their population is set to zero.

    Parameters
    ----------
    tmat : ndarray
        Quadratic transition matrix, needs to be ergodic.
    allow_non_ergodic : bool
        If True only the largest ergodic subset will be used. Otherwise it will
        throw an error if not ergodic.

    Returns
    -------
    peq : ndarray
        Equilibrium population of input matrix.

    """
    tmat = np.asarray(tmat)
    is_ergodic = tests.is_ergodic(tmat)
    if not allow_non_ergodic and not is_ergodic:
        raise ValueError('tmat needs to be ergodic transition matrix.')

    # calculate ev for ergodic subset
    if is_ergodic:
        _, eigenvectors = linalg.left_eigenvectors(tmat, nvals=1)
        eigenvectors = eigenvectors[0]
    else:
        mask = tests.ergodic_mask(tmat)
        _, evs_mask = linalg.left_eigenvectors(
            row_normalize_matrix(
                tmat[np.ix_(mask, mask)],
            ),
            nvals=1,
        )

        eigenvectors = np.zeros(len(tmat), dtype=tmat.dtype)
        eigenvectors[mask] = evs_mask[0]

    return eigenvectors / np.sum(eigenvectors)



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Plot the implied timescales."""
import click
import matplotlib.pyplot as plt
import numpy as np
import msmhelper as mh
import prettypyplot as pplt


@click.command(no_args_is_help='-h')
@click.option(
    '--filename',
    '-f',
    required=True,
    type=click.Path(exists=True),
    help='Path to state trajectory file (single column ascii file).',
)
@click.option(
    '--microfilename',
    required=False,
    type=click.Path(exists=True),
    help=(
        'Path to microstate trajectory file (single column ascii file) to use '
        'Hummer-Szabo projection.'
    ),
)
@click.option(
    '--concat-limits',
    '-c',
    required=False,
    type=click.Path(exists=True),
    help=(
        'Path to concatination limit file given the length of all ' +
        r'trajectories, e.g. "3\n3\n5"'
    ),
)
@click.option(
    '--output',
    '-o',
    type=click.Path(),
    help=(
        'Output name of figure. Needs to have a valid suffix (".pdf", ".svg", '
        '".png"). Default format is pdf.'
    ),
)
@click.option(
    '--max-lagtime',
    required=True,
    type=click.IntRange(min=0),
    help='Maximal lag time to estimate Markov state model in frames.',
)
@click.option(
    '--frames-per-unit',
    required=True,
    type=click.FLOAT,
    help='Number of frames per unit.',
)
@click.option(
    '--unit',
    required=True,
    type=click.Choice(
        ['fs', 'ps', 'ns', 'us', 'frames'],
        case_sensitive=False,
    ),
    help='Unit of data.',
)
@click.option(
    '--n-lagtimes',
    type=click.IntRange(min=1),
    default=3,
    help='Number of lagtimes to plot.',
)
@click.option(
    '--ylog',
    is_flag=True,
    help='Use logarithmic y-axis.',
)
def implied_timescales(
    filename,
    microfilename,
    concat_limits,
    output,
    max_lagtime,
    frames_per_unit,
    unit,
    n_lagtimes,
    ylog,
):
    """Estimation and visualization of the implied timescales."""
    # load file
    trajs = mh.openmicrostates(filename, limits_file=concat_limits)
    if microfilename:
        microtrajs = mh.openmicrostates(
            microfilename, limits_file=concat_limits,
        )
        trajs = mh.LumpedStateTraj(trajs, microtrajs)
    else:
        trajs = mh.StateTraj(trajs)

    # calculate implied timescales
    lagtimes = np.unique(np.linspace(1, max_lagtime, num=50, dtype=int))
    impl_times = mh.msm.implied_timescales(trajs, lagtimes=lagtimes)
    n_lagtimes = np.min((n_lagtimes, len(impl_times)))

    # setup pyplot
    kwargs = {'colors': 'pastel_autunm'}
    if n_lagtimes > 5:
        kwargs = {'colors': 'macaw', 'ncs': n_lagtimes + 1}

    pplt.use_style(figsize=2.2, **kwargs, latex=False, true_black=True)

    # plot result
    fig, ax = plt.subplots()

    for idx in range(n_lagtimes):
        pplt.plot(
            lagtimes / frames_per_unit,
            impl_times[:, idx] / frames_per_unit,
            ax=ax,
            label=idx + 1,
        )

    # set scale
    ax.set_xlim([0, max_lagtime / frames_per_unit])
    if not ylog:
        ax.set_ylim([0, ax.get_ylim()[1]])

    ax.grid(True, which='major', linestyle='--')
    ax.grid(True, which='minor', linestyle='dotted')
    ax.set_axisbelow(True)

    # show diagonal
    ax.fill_between(ax.get_xlim(), ax.get_xlim(), color='pplt:grid', zorder=0)

    # set label
    if unit == 'us':
        unit = r'\textmu{{}}s'
    ax.set_ylabel(r'time scales [{unit}]'.format(unit=unit))
    ax.set_xlabel(r'$\tau _\mathrm{{lag}}$ [{unit}]'.format(unit=unit))

    pplt.legend(
        ax=ax,
        outside='top',
        title='Implied time scale:',
    )

    if ylog:
        ax.set_yscale('log')

    if output is None:
        basename = f'{filename}.sh' if microfilename else filename
        output = f'{basename}.impltime.tmax{max_lagtime:.0f}.pdf'

    # save figure
    pplt.savefig(output)


if __name__ == '__main__':
    implied_timescales()  # pragma: no cover



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Set of filtering functions.

!!! todo
    - Correct border effects of running mean

"""
import numpy as _np
from scipy.ndimage import gaussian_filter as _gaussian_filter
from scipy.ndimage import gaussian_filter1d as _gaussian_filter_1d


def runningmean(array, window):
    r"""Compute centered running average with given window size.

    This function returns the centered based running average of the given
    data. The output of this function is of the same length as the input,
    by assuming that the given data is zero before and after the given series.
    Hence, there are border affects which are not corrected.

    !!! warning
        If the given window is even (not symmetric) it will be shifted towards
        the beginning of the current value. So for `window=4`, it will consider
        the current position \(i\), the two to the left \(i-2\) and \(i-1\) and
        one to the right \(i+1\).

    Function is taken from lapis:
    https://stackoverflow.com/questions/13728392/moving-average-or-running-mean

    Parameters
    ----------
    array : ndarray
        One dimensional numpy array.
    window : int
        Integer which specifies window-width.

    Returns
    -------
    array_rmean : ndarray
        Data which is time-averaged over the specified window.

    """
    ndim = _np.asarray(array).ndim
    if ndim > 1:
        raise ValueError(
            'Runningmean is only defined for 1D data, but'
            f'{ndim:.0f}D data were provided.'
        )

    # Calculate running mean
    return _np.convolve(
        array,
        _np.ones(window) / window,
        mode='same',
    )


def gaussian_filter(array, sigma):
    r"""Compute Gaussian filter along axis=0.

    Parameters
    ----------
    array : ndarray
        One dimensional numpy array.
    sigma : float
        Float which specifies the standard deviation of the Gaussian kernel
        (window-width).

    Returns
    -------
    array_filtered : ndarray
        Data which is time-averaged with the specified Gaussian kernel.

    """
    array = _np.asarray(array, dtype=_np.float64)
    ndim = array.ndim
    if ndim > 2:
        raise ValueError(
            'Gaussian filtering is only defined for 1D and 2D data, but'
            f'{ndim:.0f}D data were provided.'
        )

    # Calculate running mean
    if ndim == 1:
        return _gaussian_filter_1d(
            array,
            sigma=sigma,
            mode='nearest',
        )
    return _gaussian_filter(
        array,
        sigma=(sigma, 0),
        mode='nearest',
    )



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""# Set of functions for analyzing the MD trajectory.

This submodule contains methods for estimating various timescales based on a
Markov model.

"""
import random

import decorit
import numba
import numpy as np
from matplotlib.cbook import boxplot_stats

from msmhelper.md.comparison import _intersect as intersect
from msmhelper.md import estimate_paths as md_estimate_paths
from msmhelper.md import estimate_waiting_times as md_estimate_wt
from msmhelper.msm.utils import linalg
from msmhelper.utils import shift_data
from msmhelper.statetraj import StateTraj


def implied_timescales(trajs, lagtimes, ntimescales=None, reversible=False):
    r"""Calculate the implied timescales.

    Calculate the implied timescales, which are defined by

    $$t_i = - \frac{t_\text{lag}}{\log\lambda_i}$$

    the $i$-th eigenvalue $\lambda_i$.

    !!! note
        It is not checked if for higher lagtimes the dimensionality changes.

    Parameters
    ----------
    trajs : StateTraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    lagtimes : list or ndarray int
        Lagtimes for estimating the markov model given in [frames].
        This is not implemented yet!
    ntimescales : int, optional
        Number of returned lagtimes.
    reversible : bool
        If reversibility should be enforced for the markov state model.

    Returns
    -------
    ts : ndarray
        Matrix containing the implied Timescales.

    """
    # format input
    trajs = StateTraj(trajs)
    lagtimes = np.atleast_1d(lagtimes)

    # check that lag times are array of integers
    if not np.issubdtype(lagtimes.dtype, np.integer):
        raise TypeError(
            'Lagtimes needs to be integers but are {0}'.format(lagtimes.dtype),
        )
    if not (lagtimes > 0).all():
        raise TypeError('Lagtimes needs to be positive integers')
    if reversible:
        raise NotImplementedError(
            'Reversible matrices are not anymore supported.'
        )

    if ntimescales is None:
        ntimescales = trajs.nstates - 1

    # initialize result
    impl_timescales = np.zeros((len(lagtimes), ntimescales))

    for idx, lagtime in enumerate(lagtimes):
        transmat, _ = trajs.estimate_markov_model(lagtime)
        impl_timescales[idx] = _implied_timescales(
            transmat, lagtime, ntimescales=ntimescales,
        )

    return impl_timescales


def _implied_timescales(tmat, lagtime, ntimescales):
    """Calculate implied timescales.

    Parameters
    ----------
    tmat : ndarray
        Quadratic transition matrix.
    lagtime : int
        Lagtime for estimating the markov model given in [frames].
    ntimescales : int, optional
        Number of returned timescales.

    Returns
    -------
    timescales : ndarray
        Implied timescale given in frames.

    """
    tmat = np.asarray(tmat)

    eigenvalues = linalg.left_eigenvalues(tmat, nvals=ntimescales + 1)
    # for negative eigenvalues no timescale is defined
    eigenvalues[eigenvalues < 0] = np.nan
    return np.ma.divide(- lagtime, np.log(eigenvalues[1:]))


def _estimate_times(
    *,
    trajs,
    lagtime,
    start,
    final,
    steps,
    estimator,
    return_list=False,
):
    """Estimates times between stated states.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    lagtime : int
        Lag time for estimating the markov model given in [frames].
    start : int or list of
        States to start counting.
    final : int or list of
        States to start counting.
    steps : int
        Number of MCMC propagation steps of MCMC run.
    estimator : function
        Estimator to propagate mcmc and return times.
    return_list : bool
        If true a list of all events is returned, else the probability density
        together with the edges is returned.

    Returns
    -------
    ts : ndarray
        Density probability of the time distribution. If `return_list=True`,
        return a sorted (!) list containing all times.
    edges : ndarray
        Array containing the edges corresponding to the probability, given in
        frames. Only for `return_list=False`.


    """
    # check correct input format
    trajs = StateTraj(trajs)

    states_start, states_final = np.unique(start), np.unique(final)

    if intersect(states_start, states_final):
        raise ValueError('States `start` and `final` do overlap.')

    # check that all states exist in trajectory
    for states in (states_start, states_final):
        if intersect(states, trajs.states) != len(states):
            raise ValueError(
                'Selected states does not exist in state trajectory.',
            )

    # convert states to idx
    idxs_start = np.array(
        [trajs.state_to_idx(state) for state in states_start],
    )
    idxs_final = np.array(
        [trajs.state_to_idx(state) for state in states_final],
    )
    start = np.random.choice(idxs_final)

    # do not convert for pytest coverage
    if not numba.config.DISABLE_JIT:  # pragma: no cover
        idxs_start = numba.typed.List(idxs_start)
        idxs_final = numba.typed.List(idxs_final)

    # estimate cumulative transition matrix
    cummat = _get_cummat(trajs=trajs, lagtime=lagtime)

    ts = estimator(
        cummat=cummat,
        start=start,
        states_from=idxs_start,
        states_to=idxs_final,
        steps=steps,
    )
    # multiply wts by lagtime
    if return_list:
        return np.repeat(
            list(ts.keys()), list(ts.values()),
        ) * lagtime

    maxtime = max(ts.keys())
    pts = np.zeros(maxtime + 1)
    for time, count in ts.items():
        pts[time] = count
    return (
        pts / (pts.sum() * lagtime),
        np.arange(len(pts) + 1) * lagtime,
    )


@decorit.alias('estimate_wt')
def estimate_waiting_times(
    *,
    trajs,
    lagtime,
    start,
    final,
    steps,
    return_list=False,
):
    """Estimates waiting times between stated states.

    The stated states (from/to) will be treated as a basin. The function
    calculates all transitions from first entering the start-basin until first
    reaching the final-basin.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    lagtime : int
        Lag time for estimating the markov model given in [frames].
    start : int or list of
        States to start counting.
    final : int or list of
        States to start counting.
    steps : int
        Number of MCMC propagation steps of MCMC run.
    return_list : bool
        If true a list of all events is returned, else the probability density
        together with the edges is returned.

    Returns
    -------
    ts : ndarray
        Density probability of the time distribution. If `return_list=True`,
        return a sorted (!) list containing all times.
    edges : ndarray
        Array containing the edges corresponding to the probability, given in
        frames. Only for `return_list=False`.

    """
    return _estimate_times(
        trajs=trajs,
        lagtime=lagtime,
        start=start,
        final=final,
        steps=steps,
        estimator=_estimate_waiting_times,
        return_list=return_list,
    )


@decorit.alias('estimate_tt')
def estimate_transition_times(
    *,
    trajs,
    lagtime,
    start,
    final,
    steps,
    return_list=False,
):
    """Estimates transition times between stated states.

    The stated states (from/to) will be treated as a basin. The function
    calculates all transitions from leaving the start-basin until first
    reaching the final-basin.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    lagtime : int
        Lag time for estimating the markov model given in [frames].
    start : int or list of
        States to start counting.
    final : int or list of
        States to start counting.
    steps : int
        Number of MCMC propagation steps of MCMC run.
    return_list : bool
        If true a list of all events is returned, else the probability density
        together with the edges is returned.

    Returns
    -------
    ts : ndarray
        Density probability of the time distribution. If `return_list=True`,
        return a sorted (!) list containing all times.
    edges : ndarray
        Array containing the edges corresponding to the probability, given in
        frames. Only for `return_list=False`.

    """
    return _estimate_times(
        trajs=trajs,
        lagtime=lagtime,
        start=start,
        final=final,
        steps=steps,
        estimator=_estimate_transition_times,
        return_list=return_list,
    )


@numba.njit
def _propagate_MCMC_step(cummat, idx_from):
    """Propagate a single step Markov chain Monte Carlo."""
    rand = random.random()  # noqa: S311
    cummat_perm, state_perm = cummat
    cummat_perm, state_perm = cummat_perm[idx_from], state_perm[idx_from]

    for idx, cummat_idx in enumerate(cummat_perm):
        # strict less to ensure that rand=0 does not jump along unconnected
        # states with Tij=0.
        if rand < cummat_idx:
            return state_perm[idx]
    # this should never be reached, but needed for numba to ensure int return
    return len(cummat_perm) - 1


@numba.njit
def _estimate_waiting_times(
    cummat, start, states_from, states_to, steps,
):
    wts = {}

    idx_start = 0
    propagates_forwards = False
    state = start
    for idx in range(steps):
        state = _propagate_MCMC_step(cummat=cummat, idx_from=state)

        if not propagates_forwards and state in states_from:
            propagates_forwards = True
            idx_start = idx
        elif propagates_forwards and state in states_to:
            propagates_forwards = False
            wt = idx - idx_start
            if wt in wts:
                wts[wt] += 1
            else:
                wts[wt] = 1

    return wts


@numba.njit
def _estimate_transition_times(
    cummat, start, states_from, states_to, steps,
):
    tpts = {}

    idx_start = 0
    propagates_forwards = False
    state = start
    for idx in range(steps):
        state = _propagate_MCMC_step(cummat=cummat, idx_from=state)

        if state in states_from:
            propagates_forwards = True
            idx_start = idx
        elif propagates_forwards and state in states_to:
            propagates_forwards = False
            wt = idx - idx_start
            if wt in tpts:
                tpts[wt] += 1
            else:
                tpts[wt] = 1

    return tpts


def estimate_paths(
    *,
    trajs,
    lagtime,
    start,
    final,
    steps,
):
    """Estimates paths and waiting times between stated states.

    The stated states (from/to) will be treated as a basin. The function
    estimates transitions from first entering the start-basin until first
    reaching the final-basin. The results will be listed by the corresponding
    pathways, where loops are removed occuring first.

    !!! note
        This function is a simple wrapper and in contrast to
        [estimate_wt][msmhelper.msm.estimate_waiting_times] it stores the whole
        MCMC trajectory in memory. Hence, it memory-hungry.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    lagtime : int
        Lag time for estimating the markov model given in [frames].
    start : int or list of
        States to start counting.
    final : int or list of
        States to start counting.
    steps : int
        Number of MCMC propagation steps of MCMC run.

    Returns
    -------
    paths : dict
        Dictionary containing the the paths as keys and and an array holding
        the times of all paths as value.

    """
    # check correct input format
    trajs = StateTraj(trajs)

    states_start, states_final = np.unique(start), np.unique(final)

    if intersect(states_start, states_final):
        raise ValueError('States `start` and `final` do overlap.')

    # check that all states exist in trajectory
    for states in (states_start, states_final):
        if intersect(states, trajs.states) != len(states):
            raise ValueError(
                'Selected states does not exist in state trajectory.',
            )

    return md_estimate_paths(
        propagate_MCMC(trajs, lagtime, steps),
        start,
        final,
    )


def propagate_MCMC(
    trajs,
    lagtime,
    steps,
    start=-1,
):
    """Propagate Monte Carlo Markov chain.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    lagtime : int
        Lag time for estimating the markov model given in [frames].
    steps : int
        Number of MCMC propagation steps.
    start : int or list of, optional
        State to start propagating. Default (-1) is random state.

    Returns
    -------
    mcmc : ndarray
        Monte Carlo Markov chain state trajectory.

    """
    # check correct input format
    trajs = StateTraj(trajs)

    # check that all states exist in trajectory
    if start == -1:
        start = np.random.choice(trajs.states)
    elif start not in trajs.states:
        raise ValueError(
            'Selected starting state does not exist in state trajectory.',
        )

    # convert states to idx
    idx_start = trajs.state_to_idx(start)

    # estimate permuted cumulative transition matrix
    cummat = _get_cummat(trajs=trajs, lagtime=lagtime)

    # do not convert for pytest coverage
    return shift_data(
        _propagate_MCMC(  # pragma: no cover
            cummat=cummat,
            start=idx_start,
            steps=steps,
        ),
        np.arange(trajs.nstates),
        trajs.states,
    )


@numba.njit
def _propagate_MCMC(cummat, start, steps):
    mcmc = np.empty(steps, dtype=np.int32)

    state = start
    mcmc[0] = state
    for idx in range(steps - 1):
        state = _propagate_MCMC_step(cummat=cummat, idx_from=state)
        mcmc[idx + 1] = state

    return mcmc


def _get_cummat(trajs, lagtime):
    # estimate cumulative transition matrix
    msm, _ = StateTraj(trajs).estimate_markov_model(lagtime)

    if np.any(msm < 0):
        raise ValueError('An entry of T_ij is less than 0!')

    cummat_perm = np.empty_like(msm)
    state_perm = np.empty_like(msm, dtype=np.int64)

    for idx, row in enumerate(msm):
        idx_sort = np.argsort(row)[::-1]

        cummat_perm[idx] = np.cumsum(row[idx_sort])
        state_perm[idx] = idx_sort

    cummat_perm[:, -1] = 1  # enforce that probability sums up to 1
    return cummat_perm, state_perm


def _estimate_stats(coord):
    """Return boxplot stats of data."""
    Q1, Q2, Q3 = np.quantile(
        coord,
        [0.25, 0.5, 0.75],
    )
    IQR = Q3 - Q1
    return (
        coord.min(),
        max(Q1 - IQR, coord.min()),
        Q1,
        Q2,
        Q3,
        min(Q3 + IQR, coord.max()),
        coord.max()
    )


@decorit.alias('estimate_wtd')
def estimate_waiting_time_dist(
    trajs,
    max_lagtime,
    start,
    final,
    steps,
    n_lagtimes=50,
):
    """Estimate waiting time distribution.

    Parameters
    ----------
    trajs : statetraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    max_lagtime : int
        Maximal lag time for estimating the markov model given in [frames].
    start : int or list of
        States to start counting.
    final : int or list of
        States to start counting.
    steps : int
        Number of MCMC propagation steps of MCMC run.

    Returns
    -------
    wtd : dict
        Dictionary containing waiting time distribution.

    """
    lagtimes = np.unique(
        np.linspace(1, max_lagtime, num=n_lagtimes, dtype=int),
    )

    # get stats
    wtd = {
        lagtime: boxplot_stats(
            estimate_waiting_times(
                trajs=trajs,
                lagtime=lagtime,
                start=start,
                final=final,
                steps=steps,
                return_list=True,
            ),
        )[0]
        for lagtime in lagtimes
    }

    # include MD
    wtd['MD'] = boxplot_stats(
        md_estimate_wt(
            trajs=trajs,
            start=start,
            final=final,
        ),
    )
    return wtd



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Compare different state discretizations."""
import click
import msmhelper as mh
import numpy as np


@click.command(no_args_is_help='-h')
@click.option(
    '--traj1',
    required=True,
    type=click.Path(exists=True),
    help='Path to first state trajectory file (single column ascii file).',
)
@click.option(
    '--traj2',
    required=True,
    type=click.Path(exists=True),
    help='Path to second state trajectory file (single column ascii file).',
)
@click.option(
    '--method',
    type=click.Choice(['symmetric', 'directed']),
    default='symmetric',
    help='Method of calculating similarity.',
)
def compare_discretization(traj1, traj2, method):
    """Similarity measure of two different state discretizations."""
    # load files
    t1 = mh.opentxt(traj1, dtype=np.int64)
    t2 = mh.opentxt(traj2, dtype=np.int64)

    # estimate similarity measure
    sim = mh.md.compare_discretization(t1, t2, method=method)
    print(f'Similarity: {sim:.5f}\n - {traj1}\n - {traj2}\n')


if __name__ == '__main__':
    compare_discretization()



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Plot the Chapman Kolmogorov test."""
import numpy as np
import prettypyplot as pplt
from matplotlib import pyplot as plt


def plot_ck_test(
    ck,
    states=None,
    frames_per_unit=1,
    unit='frames',
    grid=(3, 3),
):
    """Plot CK-Test results.

    This routine is a basic helper function to visualize the results of
    [msmhelper.msm.chapman_kolmogorov_test][].

    Parameters
    ----------
    ck : dict
        Dictionary holding for each lagtime the CK equation and with 'md' the
        reference.
    states : ndarray, optional
        List containing all states to plot the CK-test.
    frames_per_unit : float, optional
        Number of frames per given unit. This is used to scale the axis
        accordingly.
    unit : ['frames', 'fs', 'ps', 'ns', 'us'], optional
        Unit to use for label.
    grid : (int, int), optional
        The number of `(n_rows, n_cols)` to use for the grid layout.

    Returns
    -------
    fig : matplotlib.Figure
        Figure holding plots.

    """
    # load colors
    pplt.load_cmaps()
    pplt.load_colors()

    lagtimes = np.array([key for key in ck.keys() if key != 'md'])
    if states is None:
        states = np.array(
            list(ck['md']['ck'].keys())
        )

    nrows, ncols = grid
    needed_rows = int(np.ceil(len(states) / ncols))

    fig, axs = plt.subplots(
        needed_rows,
        ncols,
        sharex=True,
        sharey='row',
        gridspec_kw={'wspace': 0, 'hspace': 0},
    )
    axs = np.atleast_2d(axs)

    max_time = np.max(ck['md']['time'])
    for irow, states_row in enumerate(_split_array(states, ncols)):
        for icol, state in enumerate(states_row):
            ax = axs[irow, icol]

            pplt.plot(
                ck['md']['time'] / frames_per_unit,
                ck['md']['ck'][state],
                '--',
                ax=ax,
                color='pplt:gray',
                label='MD',
            )
            for lagtime in lagtimes:
                pplt.plot(
                    ck[lagtime]['time'] / frames_per_unit,
                    ck[lagtime]['ck'][state],
                    ax=ax,
                    label=lagtime / frames_per_unit,
                )
            pplt.text(
                0.5,
                0.9,
                'S{0}'.format(state),
                contour=True,
                va='top',
                transform=ax.transAxes,
                ax=ax,
            )

            # set scale
            ax.set_xscale('log')
            ax.set_xlim([
                lagtimes[0] / frames_per_unit,
                max_time / frames_per_unit,
            ])
            ax.set_ylim([0, 1])
            if irow < len(axs) - 1:
                ax.set_yticks([0.5, 1])
            else:
                ax.set_yticks([0, 0.5, 1])

            ax.grid(True, which='major', linestyle='--')
            ax.grid(True, which='minor', linestyle='dotted')
            ax.set_axisbelow(True)

    # set legend
    legend_kw = {
        'outside': 'right',
        'bbox_to_anchor': (2.0, (1 - nrows), 0.2, nrows),
    } if ncols in {1, 2} else {
        'outside': 'top',
        'bbox_to_anchor': (0.0, 1.0, ncols, 0.01),
    }
    if ncols == 3:
        legend_kw['ncol'] = 3
    pplt.legend(
        ax=axs[0, 0],
        **legend_kw,
        title=fr'$\tau_\mathrm{{lag}}$ [{unit}]',
        frameon=False,
    )

    ylabel = (
        r'self-transition probability $P_{i\to i}$'
    ) if nrows >= 3 else (
        r'$P_{i\to i}$'
    )

    pplt.hide_empty_axes()
    pplt.label_outer()
    pplt.subplot_labels(
        ylabel=ylabel,
        xlabel=r'time $t$ [{unit}]'.format(unit=unit),
    )
    return fig


def _split_array(array, chunksize):
    """Split numpy array in maximal sizes of chunksize."""
    split_limits = [
        chunksize * i for i in range(1, int(len(array) / chunksize) + 1)
    ]
    split = np.split(array, split_limits)
    if not len(split[-1]):
        split = split[:-1]
    return split



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Plot the waiting time distribution."""
import numpy as np
import prettypyplot as pplt
from matplotlib import pyplot as plt


def plot_wtd(
    wtd,
    frames_per_unit=1,
    unit='frames',
    ax=None,
    show_md=True,
    show_fliers=False,
):
    """Plot waiting time distribution.

    This is a wrapper function to plot the return value of
    [msmhelper.msm.estimate_waiting_time_dist][].

    Parameters
    ----------
    wtd : dict
        Dictionary returned from `msmhelper.msm.estimate_wtd`, holding stats of
        waiting time distributions.
    frames_per_unit : float, optional
        Number of frames per given unit. This is used to scale the axis
        accordingly.
    unit : ['frames', 'fs', 'ps', 'ns', 'us'], optional
        Unit to use for label.
    ax : matplotlib.Axes, optional
        Axes to plot figure in. With `None` the current axes is used.
    show_md : bool, optional
        Include boxplot of MD data.
    show_fliers : bool, optional
        Show fliers (outliers) in MD and MSM prediction.

    Returns
    -------
    ax : matplotlib.Axes
        Return axes holding the plot.

    """
    if ax is None:
        ax = plt.gca()

    lagtimes = np.array(
        [time for time in wtd.keys() if time != 'MD'], dtype=int,
    )
    max_lagtime = lagtimes.max()

    # convert stats to array
    LB, UB, Q1, Q2, Q3 = np.array([
        np.array([
            wtd[lagtime][key] for lagtime in lagtimes
        ]) / frames_per_unit
        for key in ['whislo', 'whishi', 'q1', 'med', 'q3']
    ])
    FL = np.array([
        min(
            np.min(wtd[lagtime]['fliers']),
            wtd[lagtime]['whislo'],
        ) for lagtime in lagtimes
    ]) / frames_per_unit
    FU = np.array([
        max(
            np.max(wtd[lagtime]['fliers']),
            wtd[lagtime]['whishi'],
        ) for lagtime in lagtimes
    ]) / frames_per_unit

    # plot results
    colors = pplt.categorical_color(4, 'C0')
    lagtimes = lagtimes / frames_per_unit
    if show_fliers:
        ax.fill_between(
            lagtimes, FL, FU, color=colors[3], label=r'fliers',
        )
    ax.fill_between(
        lagtimes,
        LB,
        UB,
        color=colors[2],
        label=r'$Q_{1/3}\pm1.5\mathrm{IQR}$',
    )
    ax.fill_between(lagtimes, Q1, Q3, color=colors[1], label='IQR')
    ax.plot(lagtimes, Q2, color=colors[0], label='$Q_2$')

    max_lagtime_unit = max_lagtime / frames_per_unit
    if show_md:
        bxp = ax.bxp(
            [{
                key: time / frames_per_unit
                for key, time in wtd['MD'][0].items()
            }],
            positions=[max_lagtime_unit * 1.125],
            widths=max_lagtime_unit * 0.075,
            showfliers=show_fliers,
        )
        for median in bxp['medians']:
            median.set_color('k')

        ax.axvline(
            max_lagtime_unit,
            0,
            1,
            lw=plt.rcParams['axes.linewidth'],
            color='pplt:axes',
        )

    if show_md:
        ax.set_xlim([0, max_lagtime_unit * 1.25])
        xticks = np.array([
            *np.linspace(0, max_lagtime_unit, 4).astype(int),
            max_lagtime_unit * 1.125,
        ])
        xticklabels = [
            f'{xtick:.0f}' if idx + 1 < len(xticks) else 'MD'
            for idx, xtick in enumerate(xticks)
        ]
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels)
    else:
        ax.set_xlim([0, max_lagtime_unit])

    # use scientific notation for large values
    ax.ticklabel_format(
        axis='y', style='scientific', scilimits=[0, 2], useMathText=True,
    )
    ax.get_yaxis().get_offset_text().set_ha('right')

    # set legend and labels
    pplt.legend(ax=ax, outside='top', frameon=False)
    ax.set_ylabel(f'time $t$ [{unit}]')
    ax.set_xlabel(fr'$\tau_\mathrm{{lag}}$ [{unit}]')

    return ax



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""CLI of msmhelper.

Containing some command line interfaces to access basic functionality directly.

"""
import click

import msmhelper as mh
from msmhelper._cli.ck_test import ck_test
from msmhelper._cli.dynamical_coring import dynamical_coring
from msmhelper._cli.implied_timescales import implied_timescales
from msmhelper._cli.gaussian_filter import gaussian_filtering
from msmhelper._cli.contact_rep import contact_rep
from msmhelper._cli.compare_discretization import compare_discretization
from msmhelper._cli.waiting_time_dist import waiting_time_dist
from msmhelper._cli.waiting_times import waiting_times

HELP_STR = f"""msmhelper v{mh.__version__}

Unlock the power of protein dynamics time series with Markov state modeling, by
simplifying scientific analysis.

Copyright (c) 2019-2023, Daniel Nagel
"""


@click.group(help=HELP_STR)
def main():
    """Empty group to show on help available submodules."""
    pass


main.add_command(ck_test)
main.add_command(implied_timescales)
main.add_command(waiting_time_dist)
main.add_command(waiting_times)
main.add_command(contact_rep)
main.add_command(dynamical_coring)
main.add_command(gaussian_filtering)
main.add_command(compare_discretization)


if __name__ == '__main__':
    main()  # pragma: no cover



# -*- coding: utf-8 -*-
"""# Markov State Modeling

This submodule contains methods related to Markov state modeling, a powerful technique for analyzing complex systems. It provides a set of functions for constructing and analyzing Markov models, including methods for calculating transition probabilities and estimating various time scales.

The submodule is structured into the following submodules:

- [**msm:**][msmhelper.msm.msm] This submodule contains all methods related to estimate the Markov state model.
- [**tests:**][msmhelper.msm.tests] This submodule holds methods for validating Markov state models.
- [**timescales:**][msmhelper.msm.timescales] This submodule contains methods for estimating various timescales based on a Markov model.
- [**utils:**][msmhelper.msm.utils] This submodule provides some useful linear algebra methods.

"""
__all__ = [
    'chapman_kolmogorov_test',
    'ck_test',
    'estimate_markov_model',
    'equilibrium_population',
    'peq',
    'row_normalize_matrix',
    'implied_timescales',
    'estimate_waiting_times',
    'estimate_wt',
    'estimate_waiting_time_dist',
    'estimate_wtd',
    'estimate_paths',
]

from .tests import (
    chapman_kolmogorov_test,
    ck_test,
)
from .msm import (
    estimate_markov_model,
    equilibrium_population,
    peq,
    row_normalize_matrix,
)
from .timescales import (
    implied_timescales,
    estimate_waiting_times,
    estimate_wt,
    estimate_waiting_time_dist,
    estimate_wtd,
    estimate_paths,
)



# -*- coding: utf-8 -*-
"""Tests for the filtering submodule.

BSD 3-Clause License
Copyright (c) 2019-2023, Daniel Nagel
All rights reserved.

"""
import numpy as np
import pytest
from msmhelper.utils import filtering


@pytest.mark.parametrize('data, expected, window, error', [
    (
        [1, 1, 1, 2, 2, 2, 3, 3, 3],
        [1., 1., 1., 2., 2., 2., 3., 3., 3.],
        1,
        None,
    ),
    (
        [1, 1, 1, 2, 2, 2, 3, 3, 3],
        [.5, 1., 1., 1.5, 2., 2., 2.5, 3., 3.],
        2,
        None
    ),
    (
        np.random.normal(size=(20, 2)),
        None,
        1,
        ValueError,
    )
])
def test_runningmean(data, expected, window, error):
    """Test runningmean."""
    if error is None:
        np.testing.assert_array_almost_equal(
            filtering.runningmean(data, window),
            expected
        )
    else:
        with pytest.raises(error):
            filtering.runningmean(data, window)


@pytest.mark.parametrize('data, expected, sigma, error', [
    (
        np.arange(10),
        np.arange(10),
        0.1,
        None,
    ),
    (
        np.vstack([np.arange(10), np.zeros(10)]).T,
        np.vstack([
            filtering.gaussian_filter(np.arange(10), 2), np.zeros(10),
        ]).T,
        2,
        None,
    ),
    (
        np.arange(10).reshape(-1, 2),
        [
            [1.53241702, 2.53241702],
            [2.65493269, 3.65493269],
            [4.00000000, 5.00000000],
            [5.34506731, 6.34506731],
            [6.46758298, 7.46758298],
        ],
        2,
        None,
    ),
    (
        np.random.normal(size=(20, 5, 3)),
        None,
        1,
        ValueError,
    )
])
def test_gaussian_filter(data, expected, sigma, error):
    """Test runningmean."""
    if error is None:
        np.testing.assert_array_almost_equal(
            filtering.gaussian_filter(data, sigma),
            expected,
        )
    else:
        with pytest.raises(error):
            filtering.gaussian_filter(data, sigma)



# -*- coding: utf-8 -*-
"""Dynamical corrections.

BSD 3-Clause License
Copyright (c) 2022, Daniel Nagel
All rights reserved.

Authors: Daniel Nagel

"""
import numba

from msmhelper.statetraj import LumpedStateTraj, StateTraj


class LagtimeError(Exception):
    """An exception for the given lagtime was raised."""


def dynamical_coring(trajs, lagtime, iterative=True):
    """Fix spurious transitions with dynamical coring.

    Projecting high dimensional data onto low dimensional collective variables
    can result in spurious state transitions which can be correct for applying
    dynamical coring, for more details see Nagel et al. [^1].

    !!! note
        Applying dynamical coring on a [msmhelper.LumpedStateTraj][] is not
        supported. The reason is that while applying dynamical coring on the
        microstate level leads to different coarse-graining, applying it on the
        macrostate level the HS-Projection is not well defined anymore.

    [^1]: Nagel et al., **Dynamical coring of Markov state models**,
        *J. Chem. Phys.*, 150, 094111 (2019),
        doi:[10.1063/1.5081767](https://doi.org/10.1063/1.5081767)

    Parameters
    ----------
    trajs : StateTraj or list or ndarray or list of ndarray
        State trajectory/trajectories. The states should start from zero and
        need to be integers.
    lagtime : int
        Lagtime [frames] is the minimum time a trajectory is required to spend
        in a new state to be accepted.
    iterative : bool, optional
        If `True` dynamical coring is applied iteratively with increasing
        lagtimes, so lagtime=1, 2, ..., lagtimes.

    Returns
    -------
    trajs : StateTraj
        Dynamically corrected state trajectory.

    """
    trajs = StateTraj(trajs)

    if isinstance(trajs, LumpedStateTraj):
        raise NotImplementedError(
            'Applying dynamical coring on a LumpedStateTraj is not supported. '
            'The reason is that while applying dynamical coring on the '
            'microstate level leads to different coarse-graining, applying it '
            'on the macrostate level the HS-Projection is not well defined '
            'anymore.'
        )

    # convert trajs to numba list # noqa: SC100
    if numba.config.DISABLE_JIT:
        cored_trajs = trajs.trajs
    else:  # pragma: no cover
        cored_trajs = numba.typed.List(trajs.trajs)

    if lagtime <= 0:
        raise ValueError('The lagtime should be greater 0.')

    # if lagtime == 1 nothing changes
    if lagtime == 1:
        return trajs

    # catch if lagtime <=1
    return StateTraj(
        _dynamical_coring(cored_trajs, lagtime, iterative),
    )


@numba.njit
def _dynamical_coring(trajs, lagtime, iterative):
    """Apply dynamical coring."""
    lagtimes = (
        numba.typed.List(range(2, lagtime + 1))
    ) if iterative else (
        numba.typed.List([lagtime])
    )

    for tau in lagtimes:
        trajs = _dynamical_coring_single_lagtime(
            trajs, tau, iterative,
        )

    return trajs


@numba.njit
def _dynamical_coring_single_lagtime(trajs, lagtime, iterative):
    """Apply dynamical coring."""
    # initialize matrix
    return numba.typed.List([
        _dynamical_coring_single_traj(traj, lagtime, iterative)
        for traj in trajs
    ])


@numba.njit
def _dynamical_coring_single_traj(traj, lagtime, iterative):
    """Apply dynamical coring."""
    # initialize matrix
    core = _find_first_core(traj, lagtime)
    if core == -1:
        raise LagtimeError(
            'For the given lagtime no core can be found. '
            'Try decreasing the lagtime.'
        )

    cored_traj = traj.copy()
    for idx in range(len(traj)):  # noqa: WPS518
        if cored_traj[idx] == core:
            continue

        if _remains_in_core(idx, cored_traj, lagtime, iterative=iterative):
            core = cored_traj[idx]
        else:
            cored_traj[idx] = core
    return cored_traj


@numba.njit
def _remains_in_core(idx, traj, lagtime, iterative):
    """Remains at position idx in core for time lagtime."""
    # if remaining length is shorter than lagtime it cannot remain in the core
    if len(traj) + 1 <= idx + lagtime:
        return False

    # if applying dynamical coring iteratively, we know that every sequence is
    # at least of length lagtime - 1. Hence, it is sufficient to check for the
    # last frame.
    if iterative:
        return traj[idx] == traj[idx + lagtime - 1]
    for idxNext in range(idx + 1, idx + lagtime):
        if traj[idx] != traj[idxNext]:
            return False
    return True


@numba.njit
def _find_first_core(traj, lagtime):
    """Find first core in trajectory."""
    for idx in range(len(traj)):  # noqa: WPS518
        if _remains_in_core(idx, traj, lagtime, iterative=False):
            return traj[idx]
    return -1



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Plot the waiting time distribution."""
import click
import msmhelper as mh
import prettypyplot as pplt
from matplotlib import pyplot as plt


@click.command(no_args_is_help='-h')
@click.option(
    '--filename',
    '-f',
    required=True,
    type=click.Path(exists=True),
    help='Path to state trajectory file (single column ascii file).',
)
@click.option(
    '--microfilename',
    required=False,
    type=click.Path(exists=True),
    help=(
        'Path to microstate trajectory file (single column ascii file) to use '
        'Hummer-Szabo projection.'
    ),
)
@click.option(
    '--concat-limits',
    '-c',
    required=False,
    type=click.Path(exists=True),
    help=(
        'Path to concatination limit file given the length of all ' +
        r'trajectories, e.g. "3\n3\n5"'
    ),
)
@click.option(
    '--output',
    '-o',
    type=click.Path(),
    help=(
        'Output name of figure. Needs to have a valid suffix (".pdf", ".svg", '
        '".png"). Default format is pdf.'
    ),
)
@click.option(
    '--max-lagtime',
    required=True,
    type=click.IntRange(min=1),
    help='Maximal lag time given in frames to estimate Markov state model.',
)
@click.option(
    '--start',
    required=True,
    type=click.IntRange(min=1),
    help='State to start from.',
)
@click.option(
    '--final',
    required=True,
    type=click.IntRange(min=1),
    help='State to end in.',
)
@click.option(
    '--nsteps',
    required=True,
    type=click.IntRange(min=1),
    default=int(1e8),
    help='State to end in.',
)
@click.option(
    '--frames-per-unit',
    required=True,
    type=click.FLOAT,
    help='Number of frames per unit.',
)
@click.option(
    '--unit',
    required=True,
    type=click.Choice(
        ['fs', 'ps', 'ns', 'us', 'frames'],
        case_sensitive=False,
    ),
    help='Unit of data.',
)
def waiting_time_dist(
    filename,
    microfilename,
    concat_limits,
    output,
    max_lagtime,
    start,
    final,
    nsteps,
    frames_per_unit,
    unit,
):
    """Estimation and visualization of the waiting time distributions."""
    # setup matplotlib
    pplt.use_style(
        figsize=2.4, true_black=True, colors='pastel_autunm', latex=False,
    )

    # load file
    trajs = mh.openmicrostates(filename, limits_file=concat_limits)
    if microfilename:
        microtrajs = mh.openmicrostates(
            microfilename, limits_file=concat_limits,
        )
        trajs = mh.LumpedStateTraj(trajs, microtrajs, positive=True)
    else:
        trajs = mh.StateTraj(trajs)

    # perform test
    wtd = mh.msm.estimate_wtd(
        trajs, max_lagtime=max_lagtime, start=start, final=final, steps=nsteps,
    )

    _, ax = plt.subplots()
    mh.plot.plot_wtd(
        wtd,
        ax=ax,
        frames_per_unit=frames_per_unit,
        unit=unit,
    )

    if output is None:
        basename = f'{filename}.sh' if microfilename else filename
        output = f'{basename}.wtd.pdf'
    pplt.savefig(output)


if __name__ == '__main__':
    waiting_time_dist()  # pragma: no cover



# -*- coding: utf-8 -*-
"""Tests for the dynamical correction module.

BSD 3-Clause License
Copyright (c) 2019-2023, Daniel Nagel
All rights reserved.

"""
import numba
import numpy as np
import pytest

from msmhelper.md import corrections
from msmhelper.statetraj import LumpedStateTraj, StateTraj


@pytest.mark.parametrize('idx, traj, lagtime, iterative, result', [
    (0, [1, 1, 2, 2, 2, 2, 2, 1, 2, 0, 2, 2, 0], 2, False, True),
    (0, [1, 1, 2, 2, 2, 2, 2, 1, 2, 0, 2, 2, 0], 3, False, False),
    (0, [1, 1, 2, 2, 2, 2, 2, 1, 2, 0, 2, 2, 0], 8, False, False),
    (0, [1, 1, 2, 2, 2, 2, 2, 1, 2, 0, 2, 2, 0], 8, True, True),
    (2, [1, 1, 2, 2, 2, 2, 2, 1, 2, 0, 2, 2, 0], 3, False, True),
    (0, [1, 1, 2, 2, 2, 2, 2, 1, 2, 0, 2, 2, 0], 20, False, False),
])
def test__remains_in_core(idx, traj, lagtime, iterative, result):
    """Test finding first core."""
    # convert traj to numba # noqa: SC100
    if not numba.config.DISABLE_JIT:
        traj = numba.typed.List(traj)

    assert corrections._remains_in_core(
        idx=idx,
        traj=traj,
        lagtime=lagtime,
        iterative=iterative
    ) == result


@pytest.mark.parametrize('traj, lagtime, result', [
    ([1, 1, 2, 2, 2, 2, 2, 1, 2, 0, 2, 2, 0], 2, 1),
    ([1, 1, 2, 2, 2, 2, 2, 1, 2, 0, 2, 2, 0], 3, 2),
    ([1, 1, 2, 2, 2, 2, 2, 1, 2, 0, 2, 2, 0], 6, -1),
])
def test__find_first_core(traj, lagtime, result):
    """Test finding first core."""
    # convert traj to numba # noqa: SC100
    if not numba.config.DISABLE_JIT:
        traj = numba.typed.List(traj)

    assert corrections._find_first_core(traj, lagtime) == result


@pytest.mark.parametrize('traj, lagtime, iterative, result, error', [
    (
        [1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2, 2, 2, 3, 3],
        2,
        True,
        [1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3],
        None,
    ),
    (
        [1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2, 2, 2, 3, 3],
        3,
        True,
        [1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
        None,
    ),
    (
        [1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2, 2, 2, 3, 3],
        3,
        False,
        [1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2],
        None,
    ),
    (
        [1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2, 2, 2, 3, 3],
        4,
        False,
        [2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
        None,
    ),
    (
        [1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2, 2, 2, 3, 3],
        6,
        True,
        None,
        corrections.LagtimeError,
    ),
])
def test__dynamical_coring_single_traj(
        traj, lagtime, iterative, result, error,
):
    """Test dynamical coring of single trajectory."""
    # convert traj to numba # noqa: SC100
    if not numba.config.DISABLE_JIT:
        traj = numba.typed.List(traj)

    if error is None:
        cored_traj = corrections._dynamical_coring_single_traj(
            traj=traj,
            lagtime=lagtime,
            iterative=iterative,
        )
        np.testing.assert_array_almost_equal(cored_traj, result)
    else:
        with pytest.raises(error):
            corrections._dynamical_coring_single_traj(
                traj=traj,
                lagtime=lagtime,
                iterative=iterative,
            )


@pytest.mark.parametrize('trajs, lagtime, kwargs, result, error', [
    (
        StateTraj([1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2]),
        2,
        {},
        StateTraj([1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2]),
        None,
    ),
    (
        StateTraj([1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2]),
        3,
        {},
        StateTraj([1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2]),
        None,
    ),
    (
        StateTraj([1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2]),
        3,
        {'iterative': True},
        StateTraj([1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2]),
        None,
    ),
    (
        StateTraj([1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2]),
        3,
        {'iterative': False},
        StateTraj([1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2]),
        None,
    ),
    (
        StateTraj([1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2]),
        1,
        {},
        StateTraj([1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2]),
        None,
    ),
    (
        StateTraj([1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2]),
        0,
        {},
        None,
        ValueError,
    ),
    (
        LumpedStateTraj(
            [1, 1, 1, 2, 1, 2, 2, 1, 2, 2, 2],
            [1, 1, 3, 2, 3, 4, 2, 3, 2, 2, 4],
        ),
        2,
        {},
        None,
        NotImplementedError,
    ),
])
def test_dynamical_coring(trajs, lagtime, kwargs, result, error):
    """Test dynamical coring."""
    if error is None:
        cored_trajs = corrections.dynamical_coring(
            trajs=trajs, lagtime=lagtime, **kwargs,
        )
        assert len(cored_trajs) == len(result)
        for cored_traj, traj in zip(cored_trajs, result):
            np.testing.assert_array_almost_equal(cored_traj, traj)
    else:
        with pytest.raises(error):
            corrections.dynamical_coring(
                trajs=trajs, lagtime=lagtime, **kwargs,
            )



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
"""Dynamical coring of state trajectory."""
import click
import msmhelper as mh


@click.command(
    no_args_is_help=True,
)
@click.option(
    '-i',
    '--input',
    'input_file',
    required=True,
    type=click.Path(exists=True),
    help=(
        'Path to input file. Needs to be of shape (n_samples, n_features).'
        ' All comment lines need to start with "#"'
    ),
)
@click.option(
    '--concat-limits',
    '-c',
    required=False,
    type=click.Path(exists=True),
    help=(
        'Path to concatination limit file given the length of all ' +
        r'trajectories, e.g. "3\n3\n5"'
    ),
)
@click.option(
    '-t',
    '--tcor',
    required=True,
    type=click.IntRange(min=1),
    help='Coring window given in [frames] used for dynamical coring.',
)
@click.option(
    '-o',
    '--output',
    'output_file',
    type=click.Path(),
    help='Path to output file.',
)
def dynamical_coring(input_file, concat_limits, tcor, output_file):
    """Applying dynamical coring on state trajectory."""
    if not output_file:
        output_file = f'{input_file}.dyncor{tcor:.0f}f'
    traj = mh.openmicrostates(
        input_file, limits_file=concat_limits,
    )

    cored_traj = mh.md.dynamical_coring(traj, lagtime=tcor, iterative=True)

    mh.savetxt(
        output_file,
        cored_traj.trajs_flatten,
        header=f'Iteratively dynamical cored with tcor={tcor:.3f} frames',
        fmt='%.0f',
    )


if __name__ == '__main__':
    dynamical_coring()  # pragma: no cover



#!usr/bin/env python3
"""Plot state structure of all microstates."""
from os.path import splitext

import click
import msmhelper as mh
import numpy as np
import prettypyplot as pplt
from matplotlib import pyplot as plt
from matplotlib.cbook import boxplot_stats


@click.command(no_args_is_help='-h')
@click.option(
    '--contacts',
    'contact_file',
    required=True,
    type=click.Path(exists=True),
    help=(
        'Path to file holding all contacts (features) of shape `(n_frames, '
        'n_contacts)`.'
    ),
)
@click.option(
    '--clusters',
    'cluster_file',
    required=True,
    type=click.Path(exists=True),
    help=(
        'Path to contacts cluster file, where every row is a cluster and in '
        'each row the indices corresponding to the columns of the clusters '
        'are stated.'
    ),
)
@click.option(
    '--state',
    'state_file',
    required=True,
    type=click.Path(exists=True),
    help='Path to state trajectory.',
)
@click.option(
    '--output',
    '-o',
    type=click.Path(),
    help=(
        'Output basename of figure. Needs to have a valid extension (".pdf", '
        '".svg", ".png"). Default format is pdf.'
    ),
)
@click.option(
    '--grid',
    type=click.IntRange(min=1),
    nargs=2,
    default=(4, 3),
    help='Number of rows and cols per figure.',
)
def contact_rep(contact_file, cluster_file, state_file, output, grid):
    """Contact representation of states.

    This script creates a contact representation of states. Were the states
    are obtained by [MoSAIC](https://github.com/moldyn/MoSAIC) and the contact
    representation was introduced in Nagel et al.[^1].

    [^1]: Nagel et al., **Selecting Features for Markov Modeling: A Case Study
          on HP35.**, *J. Chem. Theory Comput.*, submitted,

    """
    # setup matplotlib
    pplt.use_style(
        figsize=0.8, colors='pastel_autunm', true_black=True, latex=False,
    )

    # load files
    state_traj = mh.opentxt(state_file, dtype=int)
    contacts = mh.opentxt(contact_file, dtype=np.float64)
    states = np.unique(state_traj)
    clusters = load_clusters(cluster_file)

    contact_idxs = np.hstack(clusters)
    n_idxs = len(contact_idxs)
    n_frames = len(contacts)

    xtickpos = np.cumsum([
        0,
        *[
            len(clust) for clust in clusters[:-1]
        ],
    ]) - 0.5
    nrows, ncols = grid
    hspace, wspace = 0, 0
    ylims = 0, np.quantile(contacts, 0.999)

    for chunk in mh.plot._ck_test._split_array(states, nrows * ncols):
        fig, axs = plt.subplots(
            int(np.ceil(len(chunk) / ncols)),
            ncols,
            sharex=True,
            sharey=True,
            squeeze=False,
            gridspec_kw={'wspace': wspace, 'hspace': hspace},
        )

        # ignore outliers
        for state, ax in zip(chunk, axs.flatten()):
            contacts_state = contacts[state_traj == state]
            pop_state = len(contacts_state) / n_frames

            # get colormap
            c1, c2, c3 = pplt.categorical_color(3, 'C0')

            stats = {
                idx: boxplot_stats(contacts_state[:, idx])[0]
                for idx in contact_idxs
            }

            for color, (key_low, key_high), label in (
                (c3, ('whislo', 'whishi'), r'$Q_{1/3} \pm 1.5\mathrm{IQR}$'),
                (c2, ('q1', 'q3'), r'$\mathrm{IQR} = Q_3 - Q_1$'),
            ):
                ymax = [stats[idx][key_high] for idx in contact_idxs]
                ymin = [stats[idx][key_low] for idx in contact_idxs]
                ax.stairs(
                    ymax,
                    np.arange(n_idxs + 1) - 0.5,
                    baseline=ymin,
                    color=color,
                    lw=0,
                    fill=True,
                    label=label,
                )

            ax.hlines(
                [stats[idx]['med'] for idx in contact_idxs],
                xmin=np.arange(n_idxs) - 0.5,
                xmax=np.arange(n_idxs) + 0.5,
                label='median',
                color=c1,
            )

            pplt.text(
                0.5,
                0.95,
                fr'S{state} {pop_state:.1%}',
                ha='center',
                va='top',
                ax=ax,
                transform=ax.transAxes,
                contour=True,
            )

            ax.set_xlim([-0.5, n_idxs - 0.5])
            ax.set_ylim(*ylims)
            ax.set_xticks(xtickpos)
            ax.set_xticklabels(np.arange(len(xtickpos)) + 1)

            ax.grid(False)
            for pos in xtickpos:
                ax.axvline(pos, color='pplt:grid', lw=1.0)

        pplt.hide_empty_axes()
        pplt.legend(
            ax=axs[0, 0],
            outside='top',
            bbox_to_anchor=(
                0,
                1.0,
                axs.shape[1] + wspace * (axs.shape[1] - 1),
                0.01,
            ),
            frameon=False,
            ncol=2,
        )
        pplt.subplot_labels(
            xlabel='contact clusters',
            ylabel='distances [nm]',
        )

        # save figure and continue
        if output is None:
            output = f'{state_file}.contactRep.pdf'
        # insert state_str between pathname and extension
        path, ext = splitext(output)
        pplt.savefig(f'{path}.state{chunk[0]:.0f}-{chunk[-1]:.0f}{ext}')


def load_clusters(filename):
    """Load clusters stored from cli.

    This method is taken from [MoSAIC](https://github.com/moldyn/MoSAIC/).

    Parameters
    ----------
    filename : str
        Filename of cluster file.

    Returns
    -------
    clusters : ndarray of shape (n_clusters, )
        A list of arrays, each containing all indices (features) for each
        cluster.

    """
    comment = '#'
    with open(filename) as clusters:
        clusters_list = [
            np.array(
                cluster.split()
            ).astype(int).tolist()
            for cluster in clusters if not cluster.startswith(comment)
        ]

    # In case of clusters of same length, numpy casted it as a 2D array.
    # To ensure that the result is an numpy array of list, we need to
    # create an empty list, adding the values in the second step
    clusters = np.empty(len(clusters_list), dtype=object)
    clusters[:] = clusters_list  # noqa: WPS362
    return clusters


if __name__ == '__main__':
    contact_rep()



# -*- coding: utf-8 -*-
"""Tests for the cli script.

BSD 3-Clause License
Copyright (c) 2019-2023, Daniel Nagel
All rights reserved.

"""
import os.path

import pytest
from click.testing import CliRunner
from msmhelper.__main__ import main

# Current directory
HERE = os.path.dirname(__file__)


def test_main():
    runner = CliRunner()
    result = runner.invoke(main)
    assert result.exit_code == 0
    assert '--help' in result.output
    assert 'Usage:' in result.output


@pytest.mark.parametrize('submodule', [
    'ck-test', 'implied-timescales', 'gaussian-filtering',
])
def test_submodules(submodule):
    runner = CliRunner()
    result = runner.invoke(main, [submodule])
    assert result.exit_code == 0
    assert '--help' in result.output
    assert 'Usage:' in result.output


def test_ck_test(tmpdir):
    runner = CliRunner()

    # create trajectories
    trajfile = 'test/assets/8state_microtraj'
    macrotrajfile = 'test/assets/8state_macrotraj'
    output = tmpdir.join('output.pdf')

    params = (
        '--lagtimes 1 2 3 4 5 --frames-per-unit 1 '
        f'--unit frames --grid 2 2 --max-time 500 -o {output}'
    )

    result = runner.invoke(
        main,
        f'ck-test {params} --filename {macrotrajfile}'.split(),
    )
    assert result.exit_code == 0

    result = runner.invoke(
        main,
        (
            f'ck-test {params} --filename {macrotrajfile} '
            f'--microfilename {trajfile}'
        ).split(),
    )
    assert result.exit_code == 0


def test_implied_timescale(tmpdir):
    runner = CliRunner()

    # create trajectories
    trajfile = 'test/assets/8state_microtraj'
    macrotrajfile = 'test/assets/8state_macrotraj'
    output = tmpdir.join('output.pdf')

    params = f'--max-lagtime 25 --frames-per-unit 1 --unit frames -o {output} '

    result = runner.invoke(
        main,
        f'implied-timescales {params} --filename {macrotrajfile}'.split(),
    )
    assert result.exit_code == 0

    result = runner.invoke(
        main,
        (
            f'implied-timescales {params} --filename {macrotrajfile} '
            f'--microfilename {trajfile}'
        ).split(),
    )
    assert result.exit_code == 0


def test_waiting_time_dist(tmpdir):
    runner = CliRunner()

    # create trajectories
    trajfile = 'test/assets/8state_microtraj'
    macrotrajfile = 'test/assets/8state_macrotraj'
    output = tmpdir.join('output.pdf')

    params = (
        f'--start 1 --final 4 --nsteps 10000 -o {output} '
        '--max-lagtime 25 --frames-per-unit 1 --unit frames'
    )

    result = runner.invoke(
        main,
        f'waiting-time-dist {params} --filename {macrotrajfile}'.split(),
    )
    assert result.exit_code == 0

    result = runner.invoke(
        main,
        (
            f'waiting-time-dist {params} --filename {macrotrajfile} '
            f'--microfilename {trajfile}'
        ).split(),
    )
    assert result.exit_code == 0


def test_waiting_times(tmpdir):
    runner = CliRunner()

    # create trajectories
    trajfile = 'test/assets/8state_microtraj'
    macrotrajfile = 'test/assets/8state_macrotraj'
    output = tmpdir.join('output.pdf')

    params = (
        f'--start 1 --final 4 --nsteps 10000 -o {output} '
        '--lagtimes 1 2 3 --frames-per-unit 1 --unit frames'
    )

    result = runner.invoke(
        main,
        f'waiting-times {params} --filename {macrotrajfile}'.split(),
    )
    assert result.exit_code == 0

    result = runner.invoke(
        main,
        (
            f'waiting-times {params} --filename {macrotrajfile} '
            f'--microfilename {trajfile}'
        ).split(),
    )
    assert result.exit_code == 0


def test_gaussian_filtering(tmpdir):
    runner = CliRunner()

    # create trajectories
    input = 'test/assets/8state_microtraj'
    output = tmpdir.join('traj.gaussian')

    params = '--sigma 2'

    result = runner.invoke(
        main,
        f'gaussian-filtering {params} -i {input} -o {output}'.split(),
    )
    assert result.exit_code == 0


def test_dynamical_coring(tmpdir):
    runner = CliRunner()

    # create trajectories
    input = 'test/assets/8state_microtraj'
    output = tmpdir.join('traj.gaussian')

    params = '--tcor 5'

    result = runner.invoke(
        main,
        f'dynamical-coring {params} -i {input} -o {output}'.split(),
    )
    assert result.exit_code == 0


def test_compare_discretization(tmpdir):
    runner = CliRunner()

    # create trajectories
    microfile = 'test/assets/8state_microtraj'
    macrofile = 'test/assets/8state_macrotraj'

    for method in ('symmetric', 'directed'):
        result = runner.invoke(
            main,
            (
                f'compare-discretization --method {method} '
                f'--traj1 {microfile} --traj2 {macrofile}'
            ).split(),
        )
        assert result.exit_code == 0



# -*- coding: utf-8 -*-
# BSD 3-Clause License
# Copyright (c) 2019-2023, Daniel Nagel
# All rights reserved.
r"""Classes for handling discrete state trajectories.

- [**StateTraj**][msmhelper.statetraj.StateTraj] is a fast implementation
  of a state trajectory and should be used for microstate dynamics.

- [**LumpedStateTraj**][msmhelper.statetraj.LumpedStateTraj] is an
  implementation of the Hummer-Szabo projection[^1] and allows to reproduce
  the microstates dynamics on the macrostates space.

!!! note
    One should also mention that for bad coarse-graining one can get negative
    entries in the transition matrix $T_{ij} < 0$. To prevent this, one can
    explicitly force $T_{ij} \ge 0$ by setting the flag `positive=True`.

[^1]: Hummer and Szabo, **Optimal Dimensionality Reduction of
      Multistate Kinetic and Markov-State Models**, *J. Phys. Chem. B*,
      119 (29), 9029-9037 (2015),
      doi: [10.1021/jp508375q](https://doi.org/10.1021/jp508375q)

"""
import numpy as np

import msmhelper as mh


class StateTraj:  # noqa: WPS214
    """Class for handling discrete state trajectories."""
    __slots__ = ('_trajs', '_states')

    def __new__(cls, trajs):
        """Initialize new instance.

        If called with instance, return instance instead.

        """
        if isinstance(trajs, StateTraj):
            return trajs
        return super().__new__(cls)

    def __init__(self, trajs):
        """Initialize StateTraj and convert to index trajectories.

        If called with StateTraj instance, it will be returned instead.

        Parameters
        ----------
        trajs : list or ndarray or list of ndarray
            State trajectory/trajectories. The states need to be integers.

        """
        if isinstance(trajs, StateTraj):
            return

        self._trajs = mh.utils.format_state_traj(trajs)

        # get number of states
        self._states = mh.utils.unique(self._trajs)

        # shift to indices
        if np.array_equal(self._states, np.arange(1, self.nstates + 1)):
            self._states = np.arange(1, self.nstates + 1)
            self._trajs = [traj - 1 for traj in self._trajs]
        elif not np.array_equal(self._states, np.arange(self.nstates)):
            self._trajs, self._states = mh.utils.rename_by_index(
                self._trajs,
                return_permutation=True,
            )

    @property
    def states(self):
        """Return active set of states.

        Returns
        -------
        states : ndarray
            Numpy array holding active set of states.

        """
        return self._states

    @property
    def nstates(self):
        """Return number of states.

        Returns
        -------
        nstates : int
            Number of states.

        """
        return len(self.states)

    @property
    def ntrajs(self):
        """Return number of trajectories.

        Returns
        -------
        ntrajs : int
            Number of trajectories.

        """
        return len(self)

    @property
    def nframes(self):
        """Return cummulative length of all trajectories.

        Returns
        -------
        nframes : int
            Number of frames of all trajectories.

        """
        return np.sum([len(traj) for traj in self._trajs])

    @property
    def trajs(self):
        """Return state trajectory.

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the input data.

        """
        if np.array_equal(self.states, np.arange(1, self.nstates + 1)):
            return [traj + 1 for traj in self._trajs]
        if np.array_equal(self.states, np.arange(self.nstates)):
            return self._trajs
        return mh.shift_data(
            self._trajs,
            np.arange(self.nstates),
            self.states,
        )

    @property
    def trajs_flatten(self):
        """Return flattened state trajectory.

        Returns
        -------
        trajs : ndarray
            1D ndarray representation of state trajectories.

        """
        return np.concatenate(self.trajs)

    @property
    def index_trajs(self):
        """Return index trajectory.

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the input data.

        """
        return self._trajs

    @property
    def index_trajs_flatten(self):
        """Return flattened index trajectory.

        Returns
        -------
        trajs : ndarray
            1D ndarray representation of index trajectories.

        """
        return np.concatenate(self.index_trajs)

    def __repr__(self):
        """Return representation of class."""
        kw = {
            'classname': self.__class__.__name__,
            'trajs': self.trajs,
        }
        return ('{classname}({trajs})'.format(**kw))

    def __str__(self):
        """Return string representation of class."""
        return ('{trajs!s}'.format(trajs=self.trajs))

    def __iter__(self):
        """Iterate over trajectories."""
        return iter(self.trajs)

    def __len__(self):
        """Return length of list of trajectories."""
        return len(self._trajs)

    def __getitem__(self, key):
        """Get key value."""
        return self.trajs.__getitem__(key)  # noqa: WPS609

    def __eq__(self, other):
        """Compare two objects."""
        if not isinstance(other, StateTraj):
            return NotImplemented
        return (
            self.ntrajs == other.ntrajs and
            all(
                np.array_equal(self[idx], other[idx])
                for idx in range(self.ntrajs)
            )
        )

    def estimate_markov_model(self, lagtime):
        """Estimates Markov State Model.

        This method estimates the MSM based on the transition count matrix.

        Parameters
        ----------
        lagtime : int
            Lag time for estimating the markov model given in [frames].

        Returns
        -------
        T : ndarray
            Transition rate matrix.

        permutation : ndarray
            Array with corresponding states.

        """
        return mh.msm.msm._estimate_markov_model(
            self.index_trajs,
            lagtime,
            self.nstates,
            self.states,
        )

    def state_to_idx(self, state):
        """Get idx corresponding to state.

        Parameters
        ----------
        state : int
            State to get idx of.

        Returns
        -------
        idx : int
            Idx corresponding to state.

        """
        idx = np.where(self.states == state)[0]
        if not idx.size:
            raise ValueError(
                'State "{state}" does not exists in trajectory.'.format(
                    state=state,
                ),
            )
        return idx[0]


class LumpedStateTraj(StateTraj):
    """Class for using the Hummer-Szabo projection with state trajectories."""
    __slots__ = (
        '_trajs',
        '_states',
        '_macrostates',
        '_state_assignment',
        'positive',
    )

    def __new__(cls, macrotrajs, microtrajs=None, positive=False):
        """Initialize new instance."""
        if isinstance(macrotrajs, LumpedStateTraj):
            return macrotrajs
        return super().__new__(cls, None)

    def __init__(self, macrotrajs, microtrajs=None, positive=False):
        r"""Initialize LumpedStateTraj.

        If called with LumpedStateTraj instance, it will be returned instead.
        This class is an implementation of the Hummer-Szabo projection[^1].

        [^1]: Hummer and Szabo, **Optimal Dimensionality Reduction of
              Multistate Kinetic and Markov-State Models**, *J. Phys. Chem. B*,
              119 (29), 9029-9037 (2015),
              doi: [10.1021/jp508375q](https://doi.org/10.1021/jp508375q)

        Parameters
        ----------
        macrotrajs : list or ndarray or list of ndarray
            Lumped state trajectory/trajectories. The states need to be
            integers and all states needs to correspond to union of
            microstates.
        microtrajs : list or ndarray or list of ndarray
            State trajectory/trajectories. EaThe states should start from zero
            and need to be integers.
        positive : bool
            If `True` $T_ij\ge0$ will be enforced, else small negative values
            are possible.

        """
        if isinstance(macrotrajs, LumpedStateTraj):
            return

        if microtrajs is None:
            raise TypeError(
                'microtrajs may only be None when macrotrajs is of type ' +
                'LumpedStateTraj.',
            )

        self.positive = positive

        # parse macrotraj
        macrotrajs = mh.utils.format_state_traj(macrotrajs)
        self._macrostates = mh.utils.unique(macrotrajs)

        # init microstate trajectories
        super().__init__(microtrajs)

        # cache flattened trajectories to speed up code for many states
        macrotrajs_flatten = np.concatenate(macrotrajs)
        microtrajs_flatten = self.microstate_trajs_flatten

        self._state_assignment = np.zeros(self.nmicrostates, dtype=np.int64)
        for idx, microstate in enumerate(self.microstates):
            idx_first = mh.utils.find_first(microstate, microtrajs_flatten)
            self._state_assignment[idx] = macrotrajs_flatten[idx_first]

    @property
    def states(self):
        """Return active set of macrostates.

        Returns
        -------
        states : ndarray
            Numpy array holding active set of states.

        """
        return self._macrostates

    @property
    def nstates(self):
        """Return number of macrostates.

        Returns
        -------
        nstates : int
            Number of states.

        """
        return len(self.states)

    @property
    def microstate_trajs(self):
        """Return microstate trajectory.

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the input data.

        """
        if np.array_equal(self.microstates, np.arange(self.nmicrostates)):
            return self._trajs
        return mh.shift_data(
            self._trajs,
            np.arange(self.nmicrostates),
            self.microstates,
        )

    @property
    def microstate_trajs_flatten(self):
        """Return flattened state trajectory.

        Returns
        -------
        trajs : ndarray
            1D ndarrays representation of state trajectories.

        """
        return np.concatenate(self.microstate_trajs)

    @property
    def microstate_index_trajs(self):
        """Return microstate index trajectory.

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the microstate index trajectory.

        """
        return self._trajs

    @property
    def microstate_index_trajs_flatten(self):
        """Return flattened microstate index trajectory.

        Returns
        -------
        trajs : ndarray
            1D ndarrays representation of microstate index trajectories.

        """
        return np.concatenate(self.microstate_index_trajs)

    @property
    def trajs(self):
        """Return macrostate trajectory.

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the input macrostate data.

        """
        return mh.shift_data(
            self._trajs,
            np.arange(self.nmicrostates),
            self._state_assignment,
        )

    @property
    def index_trajs(self):
        """Return index trajectory.

        Returns
        -------
        trajs : list of ndarrays
            List of ndarrays holding the input data.

        """
        return mh.shift_data(
            self._trajs,
            np.arange(self.nmicrostates),
            self._state_assignment_idx,
        )

    @property
    def microstates(self):
        """Return active set of microstates.

        Returns
        -------
        states : ndarray
            Numpy array holding active set of states.

        """
        return self._states

    @property
    def nmicrostates(self):
        """Return number of active set of states.

        Returns
        -------
        states : ndarray
            Numpy array holding active set of states.

        """
        return len(self.microstates)

    @property
    def state_assignment(self):
        """Return micro to macrostate assignment vector.

        Returns
        -------
        state_assignment : ndarray
            Micro to macrostate assignment vector.

        """
        return self._state_assignment

    @property
    def _state_assignment_idx(self):
        """Return micro to macrostate assignment vector.

        Returns
        -------
        state_assignment_idx : ndarray
            Micro to macrostate assignment vector.

        """
        return mh.shift_data(
            self.state_assignment,
            self.states,
            np.arange(self.nstates),
        )

    def __repr__(self):
        """Return representation of class."""
        kw = {
            'classname': self.__class__.__name__,
            'trajs': self.trajs,
            'microtrajs': self.microstate_trajs,
        }
        return ('{classname}({trajs}, {microtrajs})'.format(**kw))

    def __eq__(self, other):
        """Compare two objects."""
        if not isinstance(other, LumpedStateTraj):
            return NotImplemented
        return (
            self.ntrajs == other.ntrajs and
            all(
                np.array_equal(self.trajs[idx], other.trajs[idx])
                for idx in range(self.ntrajs)
            ) and
            np.array_equal(self.state_assignment, other.state_assignment) and
            self.positive == other.positive
        )

    def estimate_markov_model(self, lagtime):
        r"""Estimates Markov State Model.

        This method estimates the microstate MSM based on the transition count
        matrix, followed by Szabo-Hummer projection[^1] formalism to
        macrostates.

        [^1]: Hummer and Szabo, **Optimal Dimensionality Reduction of
              Multistate Kinetic and Markov-State Models**, *J. Phys. Chem. B*,
              119 (29), 9029-9037 (2015),
              doi: [10.1021/jp508375q](https://doi.org/10.1021/jp508375q)

        Parameters
        ----------
        lagtime : int
            Lag time for estimating the markov model given in [frames].

        Returns
        -------
        T : ndarray
            Transition rate matrix.

        permutation : ndarray
            Array with corresponding states.

        """
        # in the following corresponds 'i' to micro and 'a' to macro
        msm_i, _ = mh.msm.msm._estimate_markov_model(
            self.microstate_index_trajs,
            lagtime,
            self.nmicrostates,
            self.microstates,
        )
        if not mh.utils.tests.is_ergodic(msm_i):
            raise TypeError('tmat needs to be ergodic transition matrix.')
        return (self._estimate_markov_model(msm_i), self.states)

    def _estimate_markov_model(self, msm_i):
        """Estimates Markov State Model."""
        ones_i = np.ones_like(self.microstates)
        ones_a = np.ones_like(self.states)
        id_i = np.diag(ones_i)
        id_a = np.diag(ones_a)

        peq_i = mh.msm.peq(msm_i)
        peq_a = np.array([
            np.sum(peq_i[self.state_assignment == state])
            for state in self.states
        ])
        d_i = np.diag(peq_i)
        d_a = np.diag(peq_a)
        aggret = np.zeros((self.nmicrostates, self.nstates))
        aggret[(np.arange(self.nmicrostates), self._state_assignment_idx)] = 1

        m_prime = np.linalg.inv(
            id_i +
            ones_i[:, np.newaxis] * peq_i[np.newaxis:, ] -
            msm_i,
        )
        m_twoprime = np.linalg.inv(
            np.linalg.multi_dot((aggret.T, d_i, m_prime, aggret)),
        )

        msm_a = (
            id_a +
            ones_a[:, np.newaxis] * peq_a[np.newaxis:, ] -
            m_twoprime @ d_a
        )

        # enforce T_ij >= 0
        if self.positive:
            msm_a[msm_a < 0] = 0

        return mh.msm.row_normalize_matrix(msm_a)
